/* Author: Steven Ellis, 12/3/97 */
/*
 *     Copyright (C) 1997-2002 Steven Ellis <ellis@brouhaha.com>
 * 
 *     This program is free software; you can redistribute it
 *     and/or modify it under the terms of the GNU General Public
 *     License version 2 as published by the Free Software Foundation
 * 
 *     This program is distributed in the hope that it will be
 *     useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU General Public License for more
 *     details.
 * 
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
 *     02139, USA.
 */ 

/* The overall idea of this program is to use a parallellizing make and
   rsh to spread compilation jobs across multiple machines.  All of the 
   machines should have common mount points for any files that are
   needed as part of the compile.

   rsh needs to be enabled on the network, and all machines used as nc hosts
   must be reachable with rsh without entering a password.  Typically, this
   will entail creating a .rhosts file in each nc users directory listing
   all of the machines that are 'trusted' for rsh access with no password.

   When first invoked, typically as "nc make [<makeparms>]",
   nc sets up some shared memory segments for host distribution information,
   invokes make with a "-j" parameter, and begins polling live distribution
   hosts for their current load average.  When the make invocation returns,
   nc exits.

   Inside make, if a rule specifies "$(NC)" at the beginning of a command
   in a rule, then a slave nc is started.

   Slave nc's check for the shared memory segments, and choose a host to
   distribute to based on the host information in shared memory.  To send a
   job to a host, rsh is invoked, and stdio to the rsh is set up to
   pass commands to the remote machine.  The commands that are passed first
   change to the correct directory, then set some environment variables, then
   invoke the actual compilation command (which followed the "$(NC)" in the 
   rule.

   Description of the startup file format can be found in .ncrc.
   A sample Makefile (suitable for building nc itself, though building
   nc can't be parallelized since there is only 1 file) shows how
   you can optionally enable nc by scattering $(NC) around make actions.
   
   nc will avoid running jobs on a host that is listed in the startup file, if
   a file named "ncavoid" is created in the root directory of that host.
*/
/* nc features, define them to turn on the functionality */
/* RSTAT, SYSLOG */
#define RSTAT /* use rstat() to contact remote hosts, if available */
#define SYSLOG /* syslog a message as nc exits with some summary info */

#include <math.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/param.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <pwd.h>

#ifdef __linux__
/* linux doesn't seem to have the rstat() call */
#undef RSTAT
#endif

#ifdef RSTAT
#include <rpc/rpc.h>
#include <rpcsvc/rstat.h>
#endif

#ifdef SYSLOG
#include <syslog.h>
#define NC_SYSLOG_PRI (LOG_INFO | LOG_LOCAL3)
#endif

/* shared memory accessible for r/w by user and r/o by group and all
   semaphore accessible for r/a (read and alter) by user only */
#define NC_SHM_ACCESS 0644
#define NC_SEM_ACCESS 0600
#define NC_RSH           "/usr/bin/rsh"
#define NC_SH            "/bin/sh"
#define NC_UPTIME        "/usr/bin/uptime"
#if 0
#define NC_UNAME         "/bin/uname"
#else
#define NC_UNAME         "/usr/bin/uname"
#endif
#define NC_AVOID          "[ ! -f /ncavoid ]"
#define NC_CD            "cd"
#define NC_EXPORT        "export"
#define NC_LOGOUT        "exit"
#define NC_ERR_REQ       "echo NCERROR $?"
#define NC_ERR_RESP      "NCERROR"
#define NC_STAT_RESP     "OK"
#define NC_STAT_REQ      "echo OK"
#define NC_JOB_VAL       "-j"
#define NC_PROG_ENV      "NC"
#define NC_JOBS_ENV      "NC_JOBS" 
#define NC_DIR_ENV	 "NC_DIR"
#define NC_MASTER_ENV    "NC_MASTER"
#define NC_BIG_ENV       "NCBIG"
#define NC_BIG_VAL       "-b"
#define NC_QUIET_ENV     "NCQUIET"
#define NC_QUIET_VAL     "-q"
#define NC_RC_ENV        "NCRC"
#define NC_STATFILE      ".ncstat"
#define NC_RCFILE        ".ncrc"
#define NC_GLOBALRC      "/etc/ncrc"
#define NC_MASTERFILE    ".ncmaster"
/* syntax for the .ncrc file */
/* stuff for the future, configurable paths for rsh,sh,uname and uptime */
#define NC_RSH_DCL       "NCRSH"
#define NC_SH_DCL        "NCSH"
#define NC_UNAME_DCL     "NCUNAME"
#define NC_UPTIME_DCL    "NCUPTIME"
/* end of stuff for the future */
#define NC_DEFCPU_DCL    "NCDEFCPU"
#define NC_DEFJOB_DCL    "NCDEFJOB"
#define NC_TYPE_DCL      "NCTYPE"
#define NC_TRANS_DCL     "NCTRANS"
#define NC_FASTHOST_DCL  "NCFASTHOST"
#define NC_HOST_DCL      "NCHOST"
#define NC_MARGIN_DCL    "NCMARGIN"
#define NC_MAXJOBS_DCL   "NCMAXJOBS"
#define NC_ENV_DCL       "NCVAR"
/* load averages are stored as integers 8 bits shifted. */
#define NC_LOAD_SHIFT   8
#define NC_LOAD_FACTOR   (1 << NC_LOAD_SHIFT)
/* epsilon (negligible difference between loads) is 1/4 */
#define NC_LOAD_EPSILON  (1 << (NC_LOAD_SHIFT - 2))
/* high load is >2 */
#define NC_LOAD_HIGH     (2 * NC_LOAD_FACTOR)
/* low load is <3/4 */
#define NC_LOAD_LOW      (NC_LOAD_FACTOR - NC_LOAD_EPSILON)
/* really low load is <1/4 */
#define NC_LOAD_REALLY_LOW (NC_LOAD_EPSILON)
#define NC_HOST_ID       0xf00d
#define NC_CTRL_ID       0xf01d
#define NC_SEM_ID        0xf01e
#define NC_DEF_CPUS      1
#define NC_DEF_JOBS      2
#define NC_MAX_JOBS      100
#define NC_DEF_MARGIN    2
#define NC_MAX_DIGITS    4
#define NC_LINE_LEN      16384
#define NC_HOST_LEN      32
#define NC_USER_LEN      16
#define NC_TYPE_LEN      16
#define NC_VAR_LEN       64
#define NC_TRANS_LEN     64
#define NC_POLL_INTERVAL 15
#define NC_POLL_WAIT     5

#define NC_USAGE_ERR     1
#define NC_NCRC_ERR      2
#define NC_SHM_ERR       3
#define NC_TIMEOUT       4
#define NC_GEN_ERR       5
#define NC_RES_ERR       6
#define NC_DUP_ERR       7
#define NC_PATH_ERR      8

#define max(a, b) (((a) > (b)) ? (a) : (b))

static char nc_cvsid[]="$Id: nc.c,v 1.3 2003/10/04 21:56:39 ellis Exp ellis $"; 
char *hoststat[] = { "UNKNOWN", "UP", "FAILING", "DOWN", "AVOID" };

/* the following data are shared between all invocations of
   nc run from the same directory */

/* table of available host machines */

struct host {
  char name[NC_HOST_LEN];      /* name of this host */
  char status;                 /* status of this host */
#define HOST_UNKNOWN 0
#define HOST_OK      1
#define HOST_FAILING 2
#define HOST_DEAD    3
#define HOST_AVOID   4
  short flags;
#define HOST_FLAGS_NONE    0x0
#define HOST_FLAGS_FAST    0x1
  int loadavg;                 /* load average * 100 */
  unsigned short totjobs;      /* number of jobs completed */
  unsigned short bigjobs;      /* number of big jobs completed */
  unsigned short badjobs;      /* number of jobs resulting in error */
  unsigned short maxjobs;      /* maximum jobs at any one time */
  unsigned short curjobs;      /* currently assigned jobs */
  unsigned char numcpus;       /* number of processors (from .ncrc) */
  unsigned char tick;          /* ticker used for periodic polling */
  short type;                  /* "type" of this host */
  struct timeval lastjob;      /* time of last job start */
   
  int pidsoffset;              /* process ids for all the jobs on this host,
				  index into an array that starts at the
				  end of the transtab */
} *volatile hosttab;
#define ALLPIDS(hostnum) (pidstab + hosttab[hostnum].pidsoffset)
#define PIDSET(hostnum, i, val) pidstab[hosttab[hostnum].pidsoffset + i] = val
#define PIDGET(hostnum, i) pidstab[hosttab[hostnum].pidsoffset + i]
 

/* the table of environment variable to send to the remote.  Just the
   names, because the values are only obtained when the child is started,
   as they may be different then */
struct var {
  char name[NC_VAR_LEN];
} *vartab;

/* table of host types.  Usually hosttype is derived from running "uname", 
   but can be set explicitly in .ncrc */
struct type {
  char name[NC_TYPE_LEN];
} *typetab;

/* the table of 'translations' of commands from one machine type to another */
struct trans {
  short type1, type2;          /* pointers into the typetab */
  char pat1[NC_TRANS_LEN];     /* translation patterns */
  char pat2[NC_TRANS_LEN];
} *transtab;

/* the table of process ids for all hosts--each host has its own section of
   the table, starting at its pidsoffset */
int *volatile pidstab;

/* control data structure for nc */
struct ctrl {
  short ncversion;             /* Version number of NC */
#define NC_VERSION   0x7
  short flags;                 /* flags */
#define CTRL_FLAGS_NONE         0x0
#define CTRL_FLAGS_FASTHOST     0x1
  unsigned short numhosts;     /* number of hosts in hosttab */
  unsigned short numvars;      /* number of environment variables in vartab */
  unsigned short numtypes;     /* number of type elements in typetab */
  unsigned short numtrans;     /* number of translation elements in transtab */
  unsigned char numjobs;       /* maximum number of jobs to run */
  unsigned char actjobs;       /* actual number of jobs make was invoked for*/
  unsigned short nexthost;     /* the next host to select */
  struct timeval start;        /* when nc was invoked */
  struct timeval end;          /* when nc stopped (only valid from .ncstat) */
  short mastertype;            /* host type of nc master */
  unsigned char curtick;       /* current tick counter */
  char mastername[NC_HOST_LEN];/* hostname of master */
  char user[NC_USER_LEN];      /* user who launched nc */
  char startdir[MAXPATHLEN];   /* where nc was started from */
  char ncdir[MAXPATHLEN];      /* ncdir path */
  char conffile[MAXPATHLEN];   /* where config file was read from */
   
} *volatile ctrl;

/* data local to this nc invocation */

/* the current child for this invocation (slave only) */
struct stream {
  int eof;                     /* end-of-file indicator */
  int bytes;                   /* # of bytes in the buffer */
  char *ptr;                   /* pointer to start of data in buffer */
  char buf[NC_LINE_LEN];       /* buffer to store unread data on this stream*/
};
   
struct child {
  int pid;                     /* process id of the child */
  char **argv;                 /* argument vector for the child to run */
  int argc;                    /* remotely */
  int hostnum;                 /* the index into the hosttable for this one */
  int pidnum;                  /* the index into the pids for this host */
  int flags;                   /* which streams to redirect */
#define STDIN      0x1
#define STDOUT     0x2
#define STDERR     0x4
  char **env;                  /* array of env var bindings */
  int sin, sout, serr;         /* file descriptors for stdin/stdout/stderr */
  int state;                   /* the current state of this child */
  /* CD_WAIT is used by both slave and master NC, to wait for the
     result of a 'cd' to the target directory */
#define NC_STATE_CD_WAIT     0
  /* states used by slave NCs, must follow CD_WAIT numerically */
#define NC_STATE_ENV_WAIT    1  /* waiting for the result of env set up */
#define NC_STATE_EXPORT_WAIT 2  /* waiting for the result of export cmd */
#define NC_STATE_CMD_WAIT    3  /* waiting for the chosen cmd to finish */
  /* states used by the master NC for polling */
#define NC_STATE_SLEEP       4  /* delaying until time to run loadavg cmd */
#define NC_STATE_AVOID_WAIT  5  /* waiting for the "skip this host" resp */
#define NC_STATE_UNAME_WAIT  6  /* waiting for the result of "uname" command */
#define NC_STATE_LOAD_WAIT   7  /* waiting for the result of loadavg cmd */
  int errcode;
  int exitcode; 
  struct stream out, err;      /* stdin and stdout streams */
} child, *chkchildren;          /* child is either make or an rsh into
                                   a remote machine to do a job,
                                   chkchildren is only used at startup
                                   by the master nc to check for directory
                                   existence (or for load average checking
                                   if rstat() is not supported */

/* Flags from the command line */
char slaveflg, infoflg, errflg, cleanflg, verflg;
char fileflg, memflg, bigflg, quietflg, remoteflg, pathflg, jobflg;
char xflg; /* used to replace 'r' flag on remote nc runs,
              also tells us to filter some output like 'q' flag */
int debugflg = 0;
char *parg;                     /* the path flags argument */
char *marg;                     /* host type of nc master */
int jobarg;                     /* the jobs flags argument */

char pwd[FILENAME_MAX + 1];     /* the current directory */
char *argv0;                    /* the base program name */
char progname[FILENAME_MAX + 1];/* and the full path name to the program */
char *ncdir = NULL;
char ncstatfile[FILENAME_MAX + 1];  /* full pathname for the stat file */
char ncmasterfile[FILENAME_MAX + 1]; /* full pathname for the master file */

/* timestamp stuff */
unsigned char maxticks = NC_POLL_INTERVAL + 1; /* ticks wrap at this number */
struct timeval curtime;

int margin;                     /* Margin of jobs to set aside */

/* private data structures for create/destroy load/unload */

/* for the host/var tables */
key_t tabkey;
int tabid = -1;

/* for the control structure */
key_t ctrlkey;
int ctrlid = -1;

key_t semkey;
int semid = -1;

inline void debug(int level, char * format, ...) {
  va_list ap;

  if (level >= debugflg)
    return;
  va_start(ap, format);
  fprintf(stderr, "DBG[%d]:", getpid());
  vfprintf(stderr, format, ap);
  fflush(stderr);
}

void dbgperror (int level, char *message) {
  char * cp = strerror(errno);
  debug(level, "%s: %s\n", message, cp);
}

/* return the largest size that a buffer might have to be to contain
   either the input or output string of a transform (or any intermediate
   version) */
int max_transform_length(short intype, short outtype, char *input) {
  int index;
  int length = strlen(input);
   
  if (intype == outtype) return length;
   
  for (index = 0; index < ctrl->numtrans; index++) {
    char *inp = input;

    if ((intype == transtab[index].type1) && 
	(outtype == transtab[index].type2)) {
      char *match;
      if ((match = strstr(inp, transtab[index].pat1))) {
	do {
	  int len1 = strlen(transtab[index].pat1);
	  int len2 = strlen(transtab[index].pat2);
	  if (len2 > len1) 
	    length += len2 - len1;
	  inp = match + len1;
	} while ((match = strstr(inp, transtab[index].pat1)));
      }
    }
  }
  return length;
}

char *transform(short intype, short outtype, char *input, char *output, int size) {
  int index;
   
  if (intype == outtype) return input;
   
  debug(3, "Transforming '%s'\n", input);
  for (index = 0; index < ctrl->numtrans; index++) {
    int outsize = size;
    char *inp = input, *outp = output;

    if ((intype == transtab[index].type1) && 
	(outtype == transtab[index].type2)) {
      char *match;
      if ((match = strstr(inp, transtab[index].pat1))) {
	do {
	  int len = match - inp;
	  int pat2len = strlen(transtab[index].pat2);

	  if (outsize <= len + pat2len)
	    return NULL;
	  outsize -= (len + pat2len);
               
	  strncpy(outp, inp, len);
	  outp += len;
               
	  strcpy(outp, transtab[index].pat2);
	  outp += pat2len;

	  inp += len + strlen(transtab[index].pat1);
	} while ((match = strstr(inp, transtab[index].pat1)));
            
	if (outsize <= strlen(inp))
	  return NULL;

	strcpy(outp, inp);
	/* go on to next rule, flipping meaning of input and output */
	inp = input;
	input = output;
	output = inp;
      }
    }
  }
  debug(3, "Tranformed into %s\n", input);
  return input;
}

short find_type(int numtypes, char *name) {
  int index;
   
  for(index = 0; index < numtypes; index++) 
    if (strcmp(name, typetab[index].name) == 0)
      return index;
  return -1;
}

short find_host(int numhosts, char *name) {
  int index;
  for(index = 0; index < numhosts; index++) 
    if (strcmp(name, hosttab[index].name) == 0)
      return index;
  return -1;
}


int nc_shared_resource_clean() {
  if (!ctrlkey)
    ctrlkey = ftok(ncdir, NC_CTRL_ID);
  if (!tabkey)
    tabkey = ftok(ncdir, NC_HOST_ID);
  semkey = ftok(ncdir, NC_SEM_ID);
  if ((tabid = shmget(tabkey, 1, 0)) != -1)
    shmctl(tabid, IPC_RMID, 0);
  if ((ctrlid = shmget(ctrlkey, 1, 0)) != -1)
    shmctl(ctrlid, IPC_RMID, 0);
  if ((semid = semget(semkey, 1, 0)) != -1)
    semctl(semid, 0, IPC_RMID, 0);
  return 0;
}

int nc_res_err(int err, char *errstr) {
  nc_shared_resource_clean();
         
  if (!errstr)
    errstr = "";

  switch (err) {
  case EACCES:
    fprintf(stderr, "**nc** Permission denied\n");
    break;
  case ENOMEM:
    fprintf(stderr, "**nc** Out of memory\n");
    break;
  case ENOSPC:
    fprintf(stderr, "**nc** Out of space\n");
    break;
  case ENOENT:
    fprintf(stderr, "**nc** No such file or directory: %s\n", errstr);
    break;
  case EEXIST:
    fprintf(stderr, "**nc** File exists: %s\n", errstr);
    break;
  case ENAMETOOLONG:
    fprintf(stderr, "**nc** Pathname too long: %s\n", errstr);
    break;
  case ELOOP:
    fprintf(stderr, "**nc** Symbolic link loop: %s\n", errstr);
    break;
  case ENOTDIR:
    fprintf(stderr, "**nc** Not a directory: %s\n", errstr);
    break;
  default:
    fprintf(stderr, "**nc** Unknown error errno=%d %s\n", err, errstr);
  }
  exit(NC_RES_ERR);
}


int create_hosts() {
  FILE *rcfile;
  char *basename;
  char *homedir;
  char filename[FILENAME_MAX];
  int numhosts, numvars, numtypes, numtrans, numjobs, totjobs, defjobs, defcpus;
  short deftype;
  char line[NC_LINE_LEN];
  int linenum;
  char *sptr;
  int elts;
  struct sembuf buf;
  int pidsoffset = 0;
  int scan = 1;

  if (!remoteflg) {
    if (!ctrlkey)
      if ((ctrlkey = ftok(ncdir, NC_CTRL_ID)) == -1) {
        dbgperror(0, "ctrlkey ftok() failed");
        nc_res_err(errno, ncdir); 
        /* does not return */
      }
    if (!tabkey)
      if ((tabkey = ftok(ncdir, NC_HOST_ID)) == -1) {
        dbgperror(0, "tabkey ftok() failed");
        nc_res_err(errno, ncdir); 
        /* does not return */
      }
    if ((semkey = ftok(ncdir, NC_SEM_ID)) == -1) {
      dbgperror(0, "semkey ftok() failed");
      nc_res_err(errno, ncdir); 
      /* does not return */
    }
   
   
    debug(1, "ctrlkey: 0x%0.8X, tabkey: 0x%0.8X, semkey: 0x%0.8X\n", 
          ctrlkey, tabkey, semkey);
  }
  if (!(basename = getenv(NC_RC_ENV)))
    basename = NC_RCFILE;
   
  sprintf(filename, "%s/%s", pwd, basename);

  debug(2, "trying %s\n", filename);
   
  if (!(rcfile = fopen(filename, "r"))) {
    char *lastslash;

    if (!(homedir = getenv("HOME")))
      goto try_program_dir;
      
    sprintf(filename, "%s/%s", homedir, basename);

    debug(2, "trying %s\n", filename);

    if (!(rcfile = fopen(filename, "r"))) {
    try_program_dir:
      lastslash = strrchr(progname, '/');
      if (!lastslash) 
	goto try_global_file;

      sprintf(filename, "%.*s/%s", lastslash - progname, progname, 
	      basename);
      debug(2, "trying %s\n", filename);
      if (!(rcfile = fopen(filename, "r"))) {
      try_global_file:
	sprintf(filename, NC_GLOBALRC);
	debug(2, "trying %s\n", filename);
	if (!(rcfile = fopen(filename, "r")))
          return 0;
      }
    }
  }
   
  /* first count entries in file, then
     rescan the file for the information */

  numjobs = 0;
 scan_file:
  numhosts = 0;
  numvars = 0;
  numtypes = 0;
  numtrans = 0;
  totjobs = 0;
  linenum = 0;
  defjobs = NC_DEF_JOBS;
  defcpus = NC_DEF_CPUS;
  deftype = 0;

  do {
    char fast = 0;
    sptr = line;

    /* read a whole line from the .ncrc file */
    if (!fgets(line, NC_LINE_LEN, rcfile))
      break;
      
    linenum++;
      
    while (isspace(*sptr))
      sptr++;
      
    if (*sptr == '#')
      /* comment */
      continue;
    if (!strncmp(sptr, NC_ENV_DCL, sizeof(NC_ENV_DCL) - 1)) {
      if (!scan) {
	sptr += sizeof(NC_ENV_DCL);
	if ((elts = sscanf(sptr, "%s\n", vartab[numvars].name)) != 1)
	  goto syntax_error;
      }
      numvars++;
    }
    else if (!strncmp(sptr, NC_DEFJOB_DCL, sizeof(NC_DEFJOB_DCL) - 1)) {
      sptr += sizeof(NC_DEFJOB_DCL);
      if ((elts = sscanf(sptr, "%d\n", &defjobs)) != 1)
	goto syntax_error;
    }
    else if (!strncmp(sptr, NC_DEFCPU_DCL, sizeof(NC_DEFCPU_DCL) - 1)) {
      sptr += sizeof(NC_DEFCPU_DCL);
      if ((elts = sscanf(sptr, "%d\n", &defcpus)) != 1)
	goto syntax_error;
    }
    else if (!strncmp(sptr, NC_TYPE_DCL, sizeof(NC_TYPE_DCL) - 1)) {
      if (!scan) {
	char type[NC_TYPE_LEN];
         
	sptr += sizeof(NC_TYPE_DCL);
	if ((elts = sscanf(sptr, "%s\n", type)) != 1)
	  goto syntax_error;
	if (find_type(numtypes, type) >= 0) {
	  fprintf(stderr, "type %s repeated\n", type);
	  goto other_error;
	}
	strcpy(typetab[numtypes].name, type);
      }
      numtypes++;
    }
    else if (!strncmp(sptr, NC_TRANS_DCL, sizeof(NC_TRANS_DCL) - 1)) {
      if (!scan) {
	char type1[NC_TYPE_LEN];
	char type2[NC_TYPE_LEN];
            
	sptr += sizeof(NC_TRANS_DCL);
	if ((elts = sscanf(sptr, "%s %s %s %s\n", type1, type2, 
			   transtab[numtrans].pat1, 
			   transtab[numtrans].pat2)) != 4)
	  goto syntax_error;
	if ((transtab[numtrans].type1 = find_type(numtypes, type1)) < 0) {
	  fprintf(stderr, "Type %s is not a valid host type\n", type1);
	  goto other_error;
	}
	if ((transtab[numtrans].type2 = find_type(numtypes, type2)) < 0) {
	  fprintf(stderr, "Type %s is not a valid host type in\n", type2);
	  goto other_error;
	}
      }
      numtrans++;
    }
    else if (!strncmp(sptr, NC_HOST_DCL, sizeof(NC_HOST_DCL) - 1) ||
	     (!strncmp(sptr, NC_FASTHOST_DCL, sizeof(NC_FASTHOST_DCL) - 1) &&
	      (fast = 1))) {
      int jobs;
      int cpus;
      char name[NC_HOST_LEN];
      char type[NC_TYPE_LEN];
         
      if (fast)
	sptr += sizeof(NC_FASTHOST_DCL) - 1;
      else
	sptr += sizeof(NC_HOST_DCL) - 1;

      if ((elts = sscanf(sptr, "%s %d %d %s\n", name, &jobs, &cpus, type)) < 1)
	goto syntax_error;
         
      if (elts < 2)
	jobs = defjobs;
      if (elts < 3)
	cpus = defcpus;
         
      if (!scan) {
	hosttab[numhosts].flags = HOST_FLAGS_NONE;
            
	if (fast) {
	  hosttab[numhosts].flags |= HOST_FLAGS_FAST;
	  ctrl->flags |= CTRL_FLAGS_FASTHOST;
	}
	strcpy(hosttab[numhosts].name, name);
	hosttab[numhosts].status = HOST_UNKNOWN;
	if (elts == 4) {
	  if ((hosttab[numhosts].type = find_type(numtypes, type)) < 0) {
	    fprintf(stderr, "Type %s is not a valid host type in \"%s\":%d\n",
		    type, filename, linenum);
	    goto other_error;
	  }
               
	}
	else
	  /* set it when we rsh into the machine */
	  hosttab[numhosts].type = -1;
            
	hosttab[numhosts].loadavg = 0;
	hosttab[numhosts].totjobs = 0;
	hosttab[numhosts].bigjobs = 0;
	hosttab[numhosts].badjobs = 0;
	hosttab[numhosts].curjobs = 0;
	hosttab[numhosts].numcpus = cpus;
	hosttab[numhosts].lastjob.tv_sec = 0;
	hosttab[numhosts].lastjob.tv_usec = 0;
        hosttab[numhosts].pidsoffset = pidsoffset;
        if (!remoteflg) {
          hosttab[numhosts].maxjobs = jobs;
          pidsoffset += hosttab[numhosts].maxjobs;
        }
        else
          hosttab[numhosts].maxjobs = (jobs ? 1 : 0);
          
      }
      else
	numjobs += jobs;
      numhosts++;
    }
    else if (!strncmp(sptr, NC_MAXJOBS_DCL, sizeof(NC_MAXJOBS_DCL) - 1)) {
      sptr += sizeof(NC_MAXJOBS_DCL) - 1;
      if ((elts = sscanf(sptr, "%d\n", &totjobs)) != 1)
	goto syntax_error;
      if (totjobs > NC_MAX_JOBS)
	totjobs = NC_MAX_JOBS;
      else if (totjobs < 1)
	totjobs = NC_DEF_JOBS;
    }
      
    else if (!strncmp(sptr, NC_MARGIN_DCL, sizeof(NC_MARGIN_DCL) - 1)) {
      sptr += sizeof(NC_MARGIN_DCL) - 1;
      if ((elts = sscanf(sptr, "%d\n", &margin)) != 1)
	goto syntax_error;
      if (margin < 1)
	margin = 0;
    }
      
    else {
    syntax_error:
      fprintf(stderr, "Syntax error \"%s\":%d\n", filename, linenum);
    other_error:
      nc_shared_resource_clean();
      exit (NC_NCRC_ERR);
    }
  } while(!feof(rcfile));
   
  if (scan) {
    struct passwd *user = getpwuid(geteuid());
    int saved_err;
    struct utsname systeminfo;

    if (uname(&systeminfo) < 0) {
      fprintf(stderr, "Call to uname() failed\n");
      /* no need to clean, don't have any shared state yet */
      exit(NC_RES_ERR);
    }

    fseek(rcfile, 0, SEEK_SET);
      
    if (remoteflg) {
      if (!(ctrl = (struct ctrl *)malloc(sizeof(struct ctrl))))
        nc_res_err(errno, NULL);
    }
    else {
      if ((ctrlid = shmget(ctrlkey, sizeof(struct ctrl), 
                           IPC_CREAT | IPC_EXCL | NC_SHM_ACCESS)) == -1) {
        dbgperror(0, "Call to shmget(ctrlkey) failed");
        if (errno == EEXIST) {
#ifdef SYSLOG
          syslog(NC_SYSLOG_PRI,
                 "%s: errcode=%d user=%s dir=%s key=0x%X host=%s\n",
                 progname, NC_DUP_ERR, user->pw_name, ncdir, ctrlkey, 
                 systeminfo.nodename);
#endif
          fprintf(stderr, 
                  "%1$s already appears to be running here.\n"
                  "If it is not, you can run (on host %3$s):\n"
                  "\t\"%1$s -imp %2$s\"   - "
                  "to get info on the running %1$s\n"
                  "or:\n"
                  "\t\"%1$s -cp %2$s\"    - "
                  "to clean up (after a crashed %1$s)\n",
                  argv0, ncdir, systeminfo.nodename);
          
          exit(NC_DUP_ERR);
        }
        nc_res_err(errno, NULL);
      }
      
      /* once shmat has happened, anyone could get the shared memory block
         however, unless ncversion is set to NC_VERSION, the data contained
         in the shared memory block isn't valid.  Consequently, we
         don't worry about semaphore protection until after ncversion gets
         set correctly (at the end of this function) */
      if ((ctrl = (struct ctrl *)shmat(ctrlid, NULL, 0)) == (void *)-1) {
        dbgperror(0, "Call to shmat(ctrlid) failed");
        nc_res_err(errno, NULL);
      }
    }
    ctrl->ncversion = 0;
    ctrl->flags = CTRL_FLAGS_NONE;
    ctrl->numhosts = numhosts;
    ctrl->numvars = numvars;
    ctrl->numtypes = numtypes;
    ctrl->numtrans = numtrans;
    ctrl->numjobs = numjobs;
    ctrl->actjobs = 0;
    ctrl->nexthost = rand() % numhosts;
    memcpy(&ctrl->start, &curtime, sizeof(struct timeval));
    memset(&ctrl->end, 0, sizeof(struct timeval));
    ctrl->mastertype = -1;
    ctrl->curtick = 0;
    
    strcpy(ctrl->mastername, systeminfo.nodename);
    
    /* get the username */
    user = getpwuid(geteuid());
    
    strcpy(ctrl->user, user->pw_name);
    strcpy(ctrl->startdir, pwd);
    strcpy(ctrl->ncdir, ncdir);
    strcpy(ctrl->conffile, filename);
    
    if (remoteflg) {
      hosttab = (struct host *)malloc(sizeof(struct host) * numhosts);
      vartab = (struct var *)malloc(sizeof(struct var) * numvars);
      typetab = (struct type *)malloc(sizeof(struct type) * numtypes);
      transtab =  (struct trans *)malloc(sizeof(struct trans) * numtrans);
      pidstab = (int *)malloc(sizeof(int));
      /* need to check for malloc failure */
    }
    else {
      if ((tabid = shmget(tabkey, 
                          sizeof(struct host) * numhosts +
                          sizeof(struct var) * numvars +
                          sizeof(struct type) * numtypes +
                          sizeof(struct trans) * numtrans +
                          sizeof(int) * numjobs, 
                          IPC_CREAT | NC_SHM_ACCESS)) == -1) {
        dbgperror(0, "Call to shmget(tabkey) failed");
        nc_res_err(errno, NULL);
      }
      
      if ((hosttab = (struct host *)shmat(tabid, NULL, 0)) == (void *)-1) {
        dbgperror(0, "Call to shmat(tabid) failed");
        nc_res_err(errno, NULL);
      }
      vartab = (struct var *)(hosttab + numhosts);
      typetab = (struct type *)(vartab + numvars);
      transtab =  (struct trans *)(typetab + numtypes);
      pidstab = (int *)(transtab + numtrans);
      memset(pidstab + pidsoffset, 0, sizeof(int) * numjobs);
      
      if ((semid = semget(semkey, 1, 
                          IPC_CREAT | IPC_EXCL | NC_SEM_ACCESS)) == -1) {
        dbgperror(0, "Call semget() failed");
        nc_res_err(errno, NULL);
      }
      
      buf.sem_num = 0;
      buf.sem_op  = 1;
      buf.sem_flg = 0;
      
      if (semop(semid, &buf, 1)) {
        dbgperror(0, "Failed to give semaphore\n");
        nc_res_err(errno, NULL);
      }
    }
    scan = 0;
    goto scan_file;
  }

  fclose(rcfile);

  if (marg && ((ctrl->mastertype = find_type(numtypes, marg)) < 0)) {
    fprintf(stderr, "Host type %s not specified in %s\n", marg, filename);
    nc_shared_resource_clean();
    exit (NC_NCRC_ERR);
  }
    
  ctrl->actjobs = ctrl->numjobs;
   
  if (totjobs && totjobs < ctrl->actjobs)
    ctrl->actjobs = totjobs;

  if (jobflg && jobarg < ctrl->actjobs)
    ctrl->actjobs = jobarg;

  if (!margin)
    margin = NC_DEF_MARGIN;
  if (margin > ctrl->numjobs)
    if (ctrl->numjobs > NC_DEF_MARGIN)
      margin = NC_DEF_MARGIN;
    else if (ctrl->numjobs > 1)
      margin = 1;
    else
      margin = 0;

  ctrl->ncversion = NC_VERSION;
  return ctrl->numhosts;
}

int read_master() {
  char master[NC_HOST_LEN];
  FILE *masterfile;

  if (!(masterfile = fopen(ncmasterfile, "r"))) {
    dbgperror(1, "Couldn't load master name\n");
    return -1;
  }
  else {
    fread(master, 1, NC_HOST_LEN, masterfile);
    fclose(masterfile);
    return find_host(ctrl->numhosts, master);
  }
}

int read_stats() {
  FILE *statfile;

  if (!(statfile = fopen(ncstatfile, "r"))) {
    dbgperror(1, "Couldn't load statistics\n");
    return 0;
  }
  else {
    ctrl = malloc(sizeof (struct ctrl));
    fread(ctrl, sizeof(struct ctrl), 1, statfile);
    if (ctrl->ncversion != NC_VERSION) {
      fprintf(stderr, "Incorrect NC version, got %d, expecting %d\n",
	      ctrl->ncversion, NC_VERSION);
      free(ctrl);
      return 0;
    }
    hosttab = malloc(sizeof(struct host) * ctrl->numhosts);
    vartab = malloc(sizeof(struct var) * ctrl->numvars);
    typetab = malloc(sizeof(struct type) * ctrl->numtypes);
    transtab = malloc(sizeof(struct trans) * ctrl->numtrans);

    fread(hosttab, sizeof(struct host), ctrl->numhosts, statfile);
    fread(vartab, sizeof(struct var), ctrl->numvars, statfile);
    fread(typetab, sizeof(struct type), ctrl->numtypes, statfile);
    fread(transtab, sizeof(struct trans), ctrl->numtrans, statfile);
    fclose(statfile);
    return 1;
  }
}

int load_hosts(int readonly) {
  if (!ctrlkey)
    if ((ctrlkey = ftok(ncdir, NC_CTRL_ID)) == -1) {
      dbgperror(0, "ctrlkey ftok() failed");
      nc_res_err(errno, ncdir); 
      /* does not return */
    }
  if (!tabkey)
    if ((tabkey = ftok(ncdir, NC_HOST_ID)) == -1) {
      dbgperror(0, "tabkey ftok() failed");
      nc_res_err(errno, ncdir); 
      /* does not return */
    }
  if ((semkey = ftok(ncdir, NC_SEM_ID)) == -1) {
    dbgperror(0, "semkey ftok() failed");
    nc_res_err(errno, ncdir); 
    /* does not return */
  }
   
  if ((ctrlid = shmget(ctrlkey, sizeof(struct ctrl), 0)) == -1) {
      
    dbgperror(1, "Call to shmget(ctrlkey) failed");
    if (!infoflg) { 
#ifdef SYSLOG
      struct utsname systeminfo;
      char *host;
      struct passwd *user = getpwuid(geteuid());
      
      if (uname(&systeminfo) < 0) {
        fprintf(stderr, "Call to uname() failed\n");
        host = "<unknown>";
      }
      else
        host = systeminfo.nodename;
          
      syslog(NC_SYSLOG_PRI,
             "%s: errcode=%d user=%s dir=%s key=0x%X host=%s\n",
             progname, NC_SHM_ERR, user->pw_name, ncdir, ctrlkey, 
             host);
#endif
      fprintf(stderr, "**nc** BIZARRE shmget() problem\n");

      if (debugflg) {
	fprintf(stderr, "**nc** Sleeping forever: use ^C to kill\n");
        fflush(stderr);
	for(;;)
	  sleep(1);
      }
    }
    return 0;
  }
      
  if ((ctrl = 
       (struct ctrl *)shmat(ctrlid, NULL, (readonly ? 
					   SHM_RDONLY : 0))) == (void *)-1) {
    dbgperror(1, "Call to shmat(ctrlid) failed");
    return 0;
  }

  if (ctrl->ncversion != NC_VERSION) {
    fprintf(stderr, "Incorrect NC version, got %d, expecting %d\n",
	    ctrl->ncversion, NC_VERSION);
    return 0;
  }

  if ((tabid = shmget(tabkey, 
		      sizeof(struct host) * ctrl->numhosts +
		      sizeof(struct var) * ctrl->numvars + 
		      sizeof(struct type) * ctrl->numtypes +
		      sizeof(struct trans) * ctrl->numtrans  +
		      sizeof(int) * ctrl->numjobs, 0)) == -1) {
    dbgperror(1, "Call to shmget(tabkey) failed");
    return 0;
  }
   
  if ((hosttab = 
       (struct host *)shmat(tabid, NULL, (readonly ?
					  SHM_RDONLY : 0))) == (void *)-1) {
    dbgperror(1, "Call to shmat(tabid) failed");
    return 0;
  }
  vartab = (struct var *)(hosttab + ctrl->numhosts);
  typetab = (struct type *)(vartab + ctrl->numvars);
  transtab =  (struct trans *)(typetab + ctrl->numtypes);
  pidstab = (int *)(transtab + ctrl->numtrans);

  if (!readonly)
    if ((semid = semget(semkey, 1, 0)) == -1) {
      dbgperror(1, "Call to semget() failed");
      return 0;
    }
   
  return ctrl->numhosts;
}

int unload_hosts(int hostnum, int pidnum, int error) {
  if (hosttab && slaveflg && hostnum != -1) {
    struct sembuf buf;

    buf.sem_num = 0;
    buf.sem_op  = -1;
    buf.sem_flg = SEM_UNDO;
      
    semop(semid, &buf, 1);

    if (pidnum != -1)
      PIDSET(hostnum, pidnum, 0);
    hosttab[hostnum].curjobs--;
    hosttab[hostnum].totjobs++;
    if (bigflg)
      hosttab[hostnum].bigjobs++;
    if (error)
      hosttab[hostnum].badjobs++;

    buf.sem_op  = 1;

    semop(semid, &buf, 1);
  }

  if (!slaveflg && !infoflg && !cleanflg && !remoteflg) {
    if (!ctrl || ctrl->ncversion != NC_VERSION)
      fprintf(stderr, "Shared memory problem trying to save stats\n");
    else {
      FILE *statfile;
#ifdef SYSLOG
      char timebuf[64];
      long runsecs, runusecs;
      int  tindex;
      int totjobs = 0;
      int totbigjobs = 0;
      int totbadjobs = 0;
      int i;
#endif
      /* set the build elapsed time */
      gettimeofday(&ctrl->end, NULL);
         
      /* write out the file, so "nc -i" can examine it post-mortem */
      if (!(statfile = fopen(ncstatfile, "w")))
	fprintf(stderr, "Couldn't save statistics\n");
      else {
	fwrite(ctrl, sizeof(struct ctrl), 1, statfile);
	fwrite(hosttab, sizeof(struct host), ctrl->numhosts, statfile);
	fwrite(vartab, sizeof(struct var), ctrl->numvars, statfile);
	fwrite(typetab, sizeof(struct type), ctrl->numtypes, statfile);
	fwrite(transtab, sizeof(struct trans), ctrl->numtrans, statfile);
	fclose(statfile);
      }
#ifdef SYSLOG
      /* get the start time in a nice format */
      /* tindex = cftime(timebuf, "%D %T", &ctrl->start.tv_sec); */
      tindex = strftime(timebuf, sizeof(timebuf), "%D %T", 
			localtime(&ctrl->start.tv_sec));
      sprintf(timebuf + tindex, ".%04.4d", ctrl->start.tv_usec / 100);
         
      /* get the elapsed time in a nice format */
      runsecs = ctrl->end.tv_sec - ctrl->start.tv_sec;
      if (ctrl->start.tv_usec > ctrl->end.tv_usec) {
	runusecs = 1000000 - ctrl->start.tv_usec + ctrl->end.tv_usec;
	runsecs--;
      }
      else
	runusecs = ctrl->end.tv_usec - ctrl->start.tv_usec;
         
      /* get the total number of jobs */
      for(i = 0; i < ctrl->numhosts; i++) {
	totjobs += hosttab[i].totjobs;
	totbigjobs += hosttab[i].bigjobs;
	totbadjobs += hosttab[i].badjobs;
      }
      syslog(NC_SYSLOG_PRI,
	     "%s: user=%s start=%s run=%02.2d:%02.2d:%02.2d.%04.4d host=%s jobs=%d bigjobs=%d badjobs=%d\n",
	     progname, ctrl->user, timebuf,
	     runsecs / (60 * 60), (runsecs / 60) % 60,
	     runsecs % 60,  runusecs / 100,
	     ctrl->mastername, totjobs, totbigjobs, totbadjobs);
#endif
    }
  }
  
  if (ctrl)
    shmdt((void *)ctrl);
  if (hosttab)
    shmdt((void *)hosttab);

  if (!slaveflg && !infoflg) {
    if (tabid != -1)
      shmctl(tabid, IPC_RMID, 0);
    if (ctrlid != -1)
      shmctl(ctrlid, IPC_RMID, 0);
    if (semid != -1)
      semctl(semid, 0, IPC_RMID, 0);
  }
  return 0;
}

float factor(int load, int curjobs, int maxjobs, int isbig)
{
#define BIGADVANTAGE 10.0
    
  double z=((double)load / NC_LOAD_FACTOR) + (double) curjobs;
  if (isbig)
    z -= BIGADVANTAGE;
  return z;
}

int pick_host(int *pidnum, int bigjob) {
  int start, end;
  int i;
  int dobig = (bigjob && (ctrl->flags & CTRL_FLAGS_FASTHOST));
  int curhost = -1;
  int curjobs = 0;
  int lastchoice = -1;
  struct sembuf buf;

  if (!remoteflg) {
    buf.sem_num = 0;
    buf.sem_op  = -1;
    buf.sem_flg = SEM_UNDO;
    
    semop(semid, &buf, 1);
  }
  start = ctrl->nexthost;
  end = ctrl->numhosts;

 again:
  for(i = start; i < end; i++)
    if (hosttab[i].status == HOST_OK)
      if (hosttab[i].curjobs < hosttab[i].maxjobs) {
	if ((!dobig || (hosttab[i].flags & HOST_FLAGS_FAST)) &&
	    (hosttab[i].curjobs == 0) &&
	    (hosttab[i].loadavg < (dobig ? 
				   NC_LOAD_REALLY_LOW : NC_LOAD_LOW))) {
	  /* below a low threshold, if the machine has no jobs
	     then pick it, and don't look further.
	     Note that if this is a big job, we do this only
	     for FAST hosts, and then we use a really low load--the
	     idea for BIG jobs is that it is crucially important
	     to run them on a really low load host--we could pick the
	     lowest possible load, but the hysterisis effects in the
	     face of multiple independent nc's might result in poor
	     allocation */
	  curhost = i;
	  goto done;
	}
	else if (dobig && (hosttab[i].flags & HOST_FLAGS_FAST) &&
		 (curhost == -1 ||
		  !(hosttab[curhost].flags & HOST_FLAGS_FAST))) {
	  /* if the job is big and there is at least 
	     one host that support big jobs and the current
	     host is fast, and either no host has yet been selected
	     or if the selected host isn't fast, then
	     choose the fast host */
	  curhost = i;
	  curjobs = hosttab[i].curjobs;
	}
	else if (curhost == -1) {
	  /* if no host has been selected, then
	     select this host */
	  curhost = i;
	  curjobs = hosttab[i].curjobs;
	}
	else {
	  if (dobig && (hosttab[curhost].flags & HOST_FLAGS_FAST) &&
	      !(hosttab[i].flags & HOST_FLAGS_FAST)) {
	    /* if the job is big and the selected host is fast,
	       and the new host isn't fast, keep looking */
	    continue;
	  }
	  if ((hosttab[i].loadavg < NC_LOAD_HIGH * hosttab[i].numcpus) &&
	      (hosttab[i].maxjobs * curjobs >
	       hosttab[curhost].maxjobs * hosttab[i].curjobs)) {
	    /* if the load average is less than a threshold,
	       pick the host with the lowest percentage of jobs */
	    curhost = i;
	    curjobs = hosttab[i].curjobs;
	  }
	  else {
	    /* with higher load averages, pick the machine with the 
	       lowest load average,  otherwise, with the lowest
	       percentage of jobs */
	    int load_delta = hosttab[i].loadavg / hosttab[i].numcpus -
	      hosttab[curhost].loadavg / 
	      hosttab[curhost].numcpus;
	    if ((load_delta <= - NC_LOAD_EPSILON) ||
		((abs(load_delta) < NC_LOAD_EPSILON) &&
		 ((hosttab[i].maxjobs * curjobs >
		   hosttab[curhost].maxjobs * hosttab[i].curjobs) ||
		  ((hosttab[i].maxjobs * curjobs ==
		    hosttab[curhost].maxjobs * hosttab[i].curjobs) &&
		   (hosttab[i].maxjobs > hosttab[curhost].maxjobs))))) {
	      curhost = i;
	      curjobs = hosttab[i].curjobs;
	    }
	  }
	}
      }
      else if (curhost == -1 && lastchoice == -1)
	/* have at least a pick for a host that is
	   alive */
	lastchoice = i;
   
  /* if we didn't start at the beginning of the table,
     we go back to the start, make sure we don't check the
     same host twice */
  if (start != 0) {
    end = start;
    start = 0;
    goto again;
  }

  /* try to get a live one, if I couldn't find a proper one */
  if (curhost == -1)
    if (lastchoice == -1) {
      curhost = ctrl->nexthost;
      fprintf(stderr, "Using bogus choice %s\n", hosttab[curhost].name);
    }
    else {
      curhost = lastchoice;
      fprintf(stderr, "Using last choice %s\n", hosttab[curhost].name);
    }
   
 done:
  hosttab[curhost].curjobs++;

#if 0
  /* increment nexthost only if this one wasn't a 'big' job */
  if (!(dobig && (hosttab[curhost].flags & HOST_FLAGS_FAST)))
    ctrl->nexthost = (curhost + 1) % ctrl->numhosts;
#else
  /* looks like this might be a better solution, especially for sites
     that have more than 1 fast host */
  ctrl->nexthost = (curhost + 1) % ctrl->numhosts;
#endif

  for (i = 0; i < hosttab[curhost].maxjobs; i++)
    if (!PIDGET(curhost, i)) {
      PIDSET(curhost, i, -1);
      *pidnum = i;
      break;
    }
  if (i == hosttab[curhost].maxjobs)
    fprintf(stderr, "Couldn't find a spot for the pid\n");

  if (!remoteflg) {
    buf.sem_op  = 1;
    semop(semid, &buf, 1);
  }

  return curhost;
}

int absolute_path(char *path, char *cwd, char *abspath) {
  char *p = path;
  char *n;
  char *e = abspath; 

  /* if path starts with a slash, then
     we don't need the abspath */
  if (*p == '/') {
    *e = '/';
    *(e+1) = 0;
    p++;
  } 
  else {
    strcpy(abspath, cwd);
    e += strlen(abspath) - 1;
    if (*e != '/') {
      *++e = '/';
      *(e+1) = 0;
    }
  }

  /* now e will always point to the trailing slash in abspath */
   
   
  do {
    if (!(n = strchr(p, '/')))
      n = p + strlen(p);
    else
      n++;

    /* n points after next slash in p or end of p */

    if (*p == '.')
      if (!*(p+1))
	p++;
      else if (*(p+1) == '/')
	p += 2;
      else if (*(p+1) == '.')
	if (!*(p+2) || *(p+2) == '/') {
	  while(*--e != '/')
	    if (e == abspath) return 1;
	  p += (*(p+2) ? 3 : 2);
	}
    strncpy(e+1, p, n - p);
    e += n - p;
    *(e+1) = 0;
    p = n;
  } while (*p);

  /* clobber the last trailing '/' */
  if (*e == '/')
    *e = 0;

  return 0;
}

int exec_path(char *argv0, char *fullpath) {
  char *path = getenv("PATH");
  char *curdir = path;
  char *nextdir;
   
  struct stat buf;
  int bytes;
   
  if (strchr(argv0, '/')) {
    realpath(argv0, fullpath);
    return 0;
  }
   
  do {
    nextdir = strchr(curdir, ':');
    if (nextdir)
      bytes = nextdir - curdir;
    else
      bytes = strlen(curdir);
      
    sprintf(fullpath, "%.*s/%s", bytes, curdir, argv0);

    debug(3, "looking for nc in %s\n", fullpath);
      
    if (!stat(fullpath, &buf) && (buf.st_mode & S_IFREG))
      return 0;
      
  }  while (*nextdir && (curdir = nextdir + 1));
  *fullpath = 0;
  return 1;
}

int kill_slaves(int *pids, int numpids) {
  int i;
  struct sembuf buf;

  buf.sem_num = 0;
  buf.sem_op  = -1;
  buf.sem_flg = SEM_UNDO;
   
  semop(semid, &buf, 1);

  for (i = 0; i < numpids; i++)
    if (pids[i]) {
      kill(pids[i], SIGINT);
      pids[i] = 0;
    }
   
  buf.sem_op  = 1;

  semop(semid, &buf, 1);

  return 0;
}

int kill_child(struct child *child) {
  if (child->pid) {
    kill(child->pid, SIGINT);
    close(child->sin);
    close(child->sout);
    close(child->serr);
    /* Try to eliminate zombie */
    if (waitpid(child->pid, &child->exitcode, WNOHANG)==-1){
      /* Hmmm.. haven't exited yet? Then wait for a while and try again */
      sleep(1);
      waitpid(child->pid, &child->exitcode, WNOHANG);
    }
    child->pid = 0;
  }
  return 0;
}

/* spawn a child process to run argv, returning the
   child process id in *pid.  if input is passed in,
   then redirect stdin to the child from the parent, and
   return the file descriptor for input to the child in *input.
   if output is passed in, then redirect stdout to the child from
   the parent, and return the file descriptor for output from the 
   child in *output.  Returns 0 on success, negative on failure */
int spawn_child(char *argv[], struct child *child) {
  int infds[2], outfds[2], errfds[2];
  int saveout, savein, saveerr;

  /* horrible stuff to get stdout in a child to
     be available to the parent.  First, save old fd,
     then close it, then fork (for both stdin and stdout) */
  if (child->flags & STDIN) {
     if ((savein = dup(fileno(stdin))) < 0)
        return -1;

     fclose(stdin);
     if (pipe(infds))
        return -2;

     child->sin = infds[1];
  }
  else
     child->sin = -1;

  if (child->flags & STDOUT) {
     if ((saveout = dup(fileno(stdout))) < 0)
        return -3;

     fclose(stdout);
     if (pipe(outfds))
        return -4;

     child->sout = outfds[1];
  }
  else
     child->sout = -1;

  if (child->flags & STDERR) {
     if ((saveerr = dup(fileno(stderr))) < 0)
        return -5;

     fclose(stderr);
     if (pipe(errfds))
        return -6;

     child->serr = errfds[1];
  }
  else
     child->serr = -1;

  child->pid = fork();
  if (child->pid != -1 && !child->pid) { /* child */
     /* after the fork, need to close off my parents
        half of the pipe, and get a 'FILE *' on my half of the pipe,
        in order to give me a real stdin/stdout/stderr */
     if (child->flags & STDIN) {
        close(infds[1]);
        close(savein);
        if (!(stdin = fdopen(infds[0], "r")))
           return -7;

     }
     if (child->flags & STDOUT) {
        close(outfds[1]);
        close(saveout);
        if (!(stdout = fdopen(outfds[0], "w")))
           return -8;
        setvbuf(stdout, NULL, _IOLBF, 0);
     }
     if (child->flags & STDERR) {
        close(errfds[1]);
        close(saveerr);
        if (!(stderr = fdopen(errfds[0], "w")))
           return -9;
        setvbuf(stderr, NULL, _IOLBF, 0);
     }
     execvp(argv[0], argv);
     dbgperror(0, "exec() failed");
     return -10;
  }
  else { /* parent, or fork() error */
     int new;
     
     /* after the fork, need to close off the childs half of the pipe, 
        then get a 'FILE *' on the saved file descriptor
        to give me a real stdin/stdout/stderr again */
     if (child->flags & STDIN) {
        close(infds[0]);
        new = dup(savein);
        if (!(stdin = fdopen(new, "r")))
           return -11;
        close(savein);
     }
     if (child->flags & STDOUT) {
        close(outfds[0]);
        new = dup(saveout);
        if (!(stdout = fdopen(new, "w")))
           return -12;
        setvbuf(stdout, NULL, _IOLBF, 0);
        close(saveout);
     }
     if (child->flags & STDERR) {
        close(errfds[0]);
        new = dup(saveerr);
        if (!(stderr = fdopen(new, "w")))
          return -13;
        setvbuf(stderr, NULL, _IOLBF, 0);
        close(saveerr);
     }
     if (child->pid == -1) {
       perror("Call to fork() failed");
       if (child->flags & STDIN)
         close(infds[1]);
       if (child->flags & STDOUT)
         close(outfds[1]);
       if (child->flags & STDERR)
         close(errfds[1]);
       return -14;
     }
     return 0;
  }
}

int cmd_length(int argc, char **argv) {
  int len = 0;
   
  while(argc--)
    len += max_transform_length(ctrl->mastertype, 
				hosttab[child.hostnum].type,
				*argv++) + 1;
  return len;
}

char *create_cmdline(int argc, char **argv, int len) {
  char *buf1;
  char *buf2;
  char *ptr;
   
  if (!(buf1 = malloc(len)) || 
      !(buf2 = malloc(len))) {
    if (buf1) free(buf1);
    fprintf(stderr, "failed to alloc cmdline\n");
    return NULL;
  }
   
  ptr = buf1;
   
  if (argc > 0) {
    strcpy(ptr, *argv);
    argc--;
    ptr += strlen(*argv++);
    if (argc)
      while(argc--) {
	*ptr++ = ' ';
	strcpy(ptr, *argv);
	ptr += strlen(*argv++);
      }
  }
  *ptr++ = 0;

  ptr = transform(ctrl->mastertype, hosttab[child.hostnum].type, buf1, buf2,
		  len);
  if (ptr == buf1) free(buf2);
  else if (ptr == buf2) free(buf1);
  else {
    /* must be null */
    fprintf(stderr, "failed to transform cmdline\n");

    free(buf1);
    free(buf2);
  }
  return ptr;
}

/* fd is the filedescriptor to read from, buf points to the beginning
   of the character buffer.  ptr is the address of a location that
   get_line can use for temporary storage (on first call should be 
   a pointer to the start of the buffer).  bytes is the address of a 
   location that get_line can use for temp storage ( */
char *get_line(struct child *child, int streamid, int *len, int *more) {
  struct stream *stream;
  int fd;
  char *line;
  char *startaddr;
  char *endaddr;
  int bytes_read;
  int bytes_to_read;
   
  *more = 0;

  if (streamid == STDOUT) {
    stream = &(child->out);
    fd = child->sout;
  }
  else if (streamid == STDERR) {
    stream = &(child->err);
    fd = child->serr;
  }
  else {
    fprintf(stderr, "get_line passed bad streamid %d\n", streamid);
    *len = 0;
    return NULL;
  }
  startaddr = stream->ptr;

  if (stream->bytes && (line = strchr(startaddr, '\n'))) {
    *line = 0;
    debug(4, "got line: %s\n", startaddr);
    stream->ptr = line + 1;
    *len = (stream->ptr - startaddr) - 1;
    stream->bytes -= *len + 1;
    if (!stream->bytes)
      stream->ptr = stream->buf;
    return startaddr;
  }
  if (startaddr - stream->buf >= NC_LINE_LEN / 2) {
    debug(5, "moving %d bytes to start\n", stream->bytes);
    memmove(stream->buf, startaddr, stream->bytes);
    stream->ptr = stream->buf;
    startaddr = stream->buf;
  }
  endaddr = stream->ptr + stream->bytes;
  bytes_to_read = NC_LINE_LEN - (stream->ptr - stream->buf) - 
    stream->bytes - 1;

  if ((bytes_read = read(fd, endaddr, bytes_to_read)) > 0) {
    endaddr[bytes_read] = 0;
    stream->bytes += bytes_read;
    if (line = strchr(startaddr, '\n')) {
      *line = 0;
      debug(4, "got line: %s\n", startaddr);
      stream->ptr = line + 1;
      *len = stream->ptr - startaddr - 1;
      stream->bytes -= *len + 1;
      if (strchr(stream->ptr, '\n'))
        *more = 1;
      return startaddr;
    }
    else {
      debug(4, "still waiting for newline\n");
      return NULL;
         
    }
  }
  else if (bytes_read < 0)
    fprintf(stderr, "***nc***: %s read() failed", strerror(errno));

  /* assume eof if we either had an error or got no bytes */
  if (stream->bytes) {
    *len = stream->bytes;
    stream->bytes = 0;
    debug(4, "no trailing newline\n");
    stream->eof++;
    return stream->ptr;
  }
  else {
    *len = 0;
    stream->eof++;
    return NULL;
  }
}

#define MASTER_NOT_QUIET 0
#define MASTER_QUIET 1
int process_master(struct child *child, char *line, int len, int quiet, 
                   int newtick) {
  char *endstr;
  int code;
   
  if (child->state == NC_STATE_CD_WAIT) {
    if (strncmp(line, NC_ERR_RESP, sizeof(NC_ERR_RESP) - 1))
      fprintf(stdout, "%s\n", line);
    else if(code = strtol(line + sizeof(NC_ERR_RESP), &endstr, 10)) {
      fprintf(stderr, "Couldn't cd to %s on %s\n",
	      pwd, hosttab[child->hostnum].name);
      hosttab[child->hostnum].status = HOST_DEAD;
      kill_child(child);
      return 1;
    }
    else {
      int cmdlen = sizeof(NC_AVOID) + sizeof(NC_ERR_REQ) + 2;
      char *cmd = alloca(cmdlen);

      strcpy(cmd, NC_AVOID "; " NC_ERR_REQ "\n");
      debug(2, "master sending %s command: %s", 
	    hosttab[child->hostnum].name, cmd);
      write(child->sin, cmd, strlen(cmd));
      child->state = NC_STATE_AVOID_WAIT;
    }
  }
  else if (child->state == NC_STATE_AVOID_WAIT) {
    if (strncmp(line, NC_ERR_RESP, sizeof(NC_ERR_RESP) - 1))
      fprintf(stdout, "%s\n", line);
    else if(code = strtol(line + sizeof(NC_ERR_RESP), &endstr, 10)) {
      fprintf(stderr, "Avoiding %s\n", hosttab[child->hostnum].name);
      hosttab[child->hostnum].status = HOST_AVOID;
      kill_child(child);
      return 1;
    }
    else
      if ((hosttab[child->hostnum].type == -1) && ctrl->numtrans) {
	int cmdlen = sizeof(NC_UNAME) + 1;
	char *cmd = alloca(cmdlen);
            
	strcpy(cmd, NC_UNAME "\n");
	debug(2, "master sending %s command: %s", 
	      hosttab[child->hostnum].name, cmd);
	write(child->sin, cmd, strlen(cmd));
	child->state = NC_STATE_UNAME_WAIT;
      }
      else 
	send_load:
      {
	int cmdlen = sizeof(NC_UPTIME) + 1;
	char *cmd = alloca(cmdlen);
#ifdef RSTAT
	struct statstime stat;
            
	if (rstat(hosttab[child->hostnum].name, &stat) == RPC_SUCCESS) {
	  hosttab[child->hostnum].status = HOST_OK;
	  hosttab[child->hostnum].loadavg = stat.avenrun[0];
	  debug(2, "%s speaks rstat\n", 
		hosttab[child->hostnum].name);
	  if (!(quiet || xflg))
	    printf("%s load: %4.2f (via rstat)\n", 
		   hosttab[child->hostnum].name,
		   (double)hosttab[child->hostnum].loadavg / 
		   NC_LOAD_FACTOR);
	  kill_child(child);
	  return 0;
	}
#endif
	strcpy(cmd, NC_UPTIME "\n");
	debug(2, "master sending %s command: %s", 
	      hosttab[child->hostnum].name, cmd);
	write(child->sin, cmd, strlen(cmd));
	child->state = NC_STATE_LOAD_WAIT;
      }
  }
  else if (child->state == NC_STATE_UNAME_WAIT) {
    if ((hosttab[child->hostnum].type = find_type(ctrl->numtypes, line)) < 0) {
      fprintf(stderr, "Couldn't find type %s from host %s\n", line, 
	      hosttab[child->hostnum].name);
      hosttab[child->hostnum].status = HOST_DEAD;
      kill_child(child);
      return 1;
    }
    goto send_load;
  }
  else if (child->state == NC_STATE_LOAD_WAIT) {
    char *loadstr;
    double load;

    loadstr = strrchr(line, ':');
    if (loadstr && ((load = strtod(++loadstr, &endstr)) || 
		    loadstr != endstr)) {
      hosttab[child->hostnum].loadavg = rint(load * NC_LOAD_FACTOR);
      if (!(quiet || xflg)) 
	printf("%s load: %4.2f\n", 
	       hosttab[child->hostnum].name,
	       (double)hosttab[child->hostnum].loadavg / NC_LOAD_FACTOR);
      else
	debug(2, "got load average from %s:%4.2f\n",
	      hosttab[child->hostnum].name,
	      (double)hosttab[child->hostnum].loadavg / NC_LOAD_FACTOR);

      hosttab[child->hostnum].status = HOST_OK;
      child->state = NC_STATE_SLEEP;
      hosttab[child->hostnum].tick = (newtick + NC_POLL_INTERVAL) % 
	maxticks;
    }
    else {
      fprintf(stderr, "couldn't get load average from %s\n",
	      hosttab[child->hostnum].name);
      hosttab[child->hostnum].status = HOST_DEAD;
      kill_slaves(ALLPIDS(child->hostnum),
		  hosttab[child->hostnum].maxjobs);
      kill_child(child);
      return 1;
    }
  }
  else {
    fprintf(stderr, "not expecting output from %s: %s\n",
	    hosttab[child->hostnum].name, line);
    return 1;
  }
  return 0;
}

/* the nc slave command--want to filter this out of make output, iff it
   appears--if we're not doing nc (distribute failed), then cmdstr is
   empty */
char *cmdstr;
int cmdstrlen = 0;

inline int process_cmd(struct child *child, int streamid, char *line, 
                       int len) {
  char *loadstr, *endstr;
  double load;
  char *strname;
  FILE *str;

  switch(streamid) {
  case STDOUT:
    strname = "stdout";
    str = stdout;
    break;
  case STDERR:
    strname = "stderr";
    str = stderr;
    break;
  default:
    nc_shared_resource_clean();
    exit(NC_GEN_ERR);
  }
  
  debug(3, "%s line was '%s'\n", strname, line);

  if (cmdstr == NULL || strncmp(cmdstr, line, cmdstrlen))
    fprintf(str, "%s\n", line);
  else
    debug(3, "killed output: %s\n", line);
  return 0;
}

int process_slave(struct child *child, char *line, int len) {
  char *endstr;
  char *cmd, *cmdptr;
  int cmdlen;
  int i;

  debug(2, "DEBUG process_slave: %d %s %d\n", child->state, line, 
	len);
   
  switch(child->state) {
  case NC_STATE_CD_WAIT:
    if (len != 2 || strcmp(line, NC_STAT_RESP))
      fprintf(stderr, "bad status in state %d from %s: %s\n", child->state, 
	      hosttab[child->hostnum].name, line);

    /* add up vars and send them */
    if (ctrl->numvars) {
      int found = 0;
      cmdlen = sizeof(NC_STAT_REQ) + 3;
         
      child->env = malloc(sizeof (char *) * ctrl->numvars);
      for (i = 0; i < ctrl->numvars; i++) {
	char *ptr;
	if (ptr = getenv(vartab[i].name)) {
	  found = 1;
	  if (hosttab[child->hostnum].type != ctrl->mastertype) {
	    int xlen = max_transform_length(ctrl->mastertype,
					    hosttab[child->hostnum].type,
					    ptr);
	    char *inbuf, *outbuf;
                  
	    if (!(inbuf = malloc(xlen)) ||
		!(outbuf = malloc(xlen))) {
	      if (inbuf) free(inbuf);
	      fprintf(stderr, "failed to alloc space for var %s\n",
		      vartab[i].name);
	      goto err_exit;
	    }

	    strcpy(inbuf, ptr);
                  
	    ptr = transform(ctrl->mastertype, 
			    hosttab[child->hostnum].type,
			    inbuf, outbuf, xlen);
	    if (!ptr) {
	      fprintf(stderr, "failed to transform var %s\n",
		      vartab[i].name);
	      free(inbuf);
	      free(outbuf);
	      goto err_exit;
	    }
	    if (ptr == inbuf)
	      free(outbuf);
	    else 
	      free(inbuf);

	    child->env[i] = ptr;
	  }
	  else
	    child->env[i] = strdup(ptr);

	  cmdlen += strlen(vartab[i].name) + strlen(child->env[i]) + 4;
	}
	else
	  child->env[i] = NULL;
      }

      if (found) {
	cmd = alloca(cmdlen);
	cmdptr = cmd;
	for (i = 0; i < ctrl->numvars; i++) {
	  if (child->env[i]) {
	    cmdptr += sprintf(cmdptr, "%s='%s' ",
			      vartab[i].name, child->env[i]);
	  }
	}
	cmdptr += sprintf(cmdptr, "; " NC_STAT_REQ "\n");
	debug(2, "sending %s command: %s", 
	      hosttab[child->hostnum].name, cmd);
	write(child->sin, cmd, strlen(cmd));
      }
      else {
	free(child->env);
	child->env = NULL;
	child->state = NC_STATE_EXPORT_WAIT;
	goto send_cmd;
      }
    }
    else {
      child->state = NC_STATE_EXPORT_WAIT;
      goto send_cmd;
    }
    child->state++;
    break;
      
  case NC_STATE_ENV_WAIT:
    if (len != 2 || strcmp(line, NC_STAT_RESP))
      fprintf(stderr, "bad status in state %d from %s: %s\n", child->state, 
	      hosttab[child->hostnum].name, line);

    /* add up export cmd, send it */
    cmdlen = sizeof(NC_EXPORT) + sizeof(NC_STAT_REQ) + 2;
      
    for (i = 0; i < ctrl->numvars; i++)
      if (child->env[i])
	cmdlen += strlen(vartab[i].name) + 1;
      
    cmd = alloca(cmdlen);
    cmdptr = cmd;
    cmdptr += sprintf(cmdptr, NC_EXPORT " ");
    for (i = 0; i < ctrl->numvars; i++) {
      if (child->env[i]) {
	cmdptr += sprintf(cmdptr, "%s ", vartab[i].name);
	free(child->env[i]);
	child->env[i] = NULL;
      }
    }
    cmdptr--;
    cmdptr += sprintf(cmdptr, "; " NC_STAT_REQ "\n");
    debug(2, "sending %s command: %s", hosttab[child->hostnum].name, cmd);
    write(child->sin, cmd, strlen(cmd));
    free(child->env);
    child->env = NULL;
    child->state++;
    break;
      
  case NC_STATE_EXPORT_WAIT: 
    {
      if (len != 2 || strcmp(line, NC_STAT_RESP))
	fprintf(stderr, "bad status in state %d from %s: %s\n", child->state, 
		hosttab[child->hostnum].name, line);
    send_cmd:
      /* send real cmd */
      
      cmd = create_cmdline(child->argc, child->argv, 
                           cmd_length(child->argc, child->argv) +
                           sizeof(NC_ERR_REQ) + sizeof(NC_LOGOUT) + 4);
      
      if (remoteflg)
	fprintf(stdout, "On %s, ", hosttab[child->hostnum].name);
      else if (!quietflg)
	fprintf(stdout, "%s:%s\n", hosttab[child->hostnum].name, cmd);
      
      cmdptr = cmd + strlen(cmd);
      cmdptr += sprintf(cmdptr, "; " NC_ERR_REQ "; " NC_LOGOUT "\n");
      debug(2, "sending %s command: %s", 
            hosttab[child->hostnum].name, cmd);
      write(child->sin, cmd, strlen(cmd));

      child->state++;
      break;
    }
      
  case NC_STATE_CMD_WAIT:
    if (strncmp(line, NC_ERR_RESP, sizeof(NC_ERR_RESP) - 1))
      fprintf(stdout, "%s\n", line);
    else
      child->errcode = strtol(line + sizeof(NC_ERR_RESP), &endstr, 10);
    break;
  default:
  err_exit:
    fprintf(stderr, "process_slave error %d\n", child->state);
    return -1;
  }
  return 0;
}

void handle_sig(int sig) {
  int i;
   
  if (slaveflg || infoflg || cleanflg) {
    if (slaveflg && child.pid) {
      kill_child(&child);
      unload_hosts(child.hostnum, child.pidnum, 1);
    }
    else
      unload_hosts(-1, 0, 1);
  }
  else {
    /* need to kill off stranded processes?? */
    kill_child(&child);
    if (ctrl && ctrl != (void *)-1 && chkchildren) 
      for (i = 0; i < ctrl->numhosts; i++)
	kill_child(&(chkchildren[i]));
    unload_hosts(-1, 0, 1);
  }
  exit(1);
}

int nc_clean() {
  if (!fileflg) {
#ifdef SYSLOG
    struct utsname systeminfo;
    char *host;
    struct passwd *user = getpwuid(geteuid());
          
    if (uname(&systeminfo) < 0) {
      fprintf(stderr, "Call to uname() failed\n");
      host = "<unknown>";
    }
    else
      host = systeminfo.nodename;

    if (!ctrlkey)
      ctrlkey = ftok(ncdir, NC_CTRL_ID);

    syslog(NC_SYSLOG_PRI,
           "%s: status=clean user=%s dir=%s key=0x%X host=%s\n",
           progname, user->pw_name, ncdir, ctrlkey, host);
#endif
    nc_shared_resource_clean();
  }
  if (!memflg)
    unlink(ncstatfile);
  return 0;
}

int nc_info() {
  int i, j;
  int err = 0;
  char *ldr;
  int cumloadavg = 0, cumtotjobs = 0, cumcurjobs = 0;
  int cumbigjobs = 0, cumbadjobs = 0;
  int cumcpus = 0;
  long secs, usecs;
  long lastsecs = 0, lastusecs = 0;
  long cursecs, curusecs;
  int curtick;
  char timebuf[32];
  struct timeval curtime;
  int  tindex;
  int  column;
  int  file = 0;
   
  if (fileflg || !load_hosts(1)) {
    if (memflg || !read_stats()) {
      fprintf(stderr, "%s does not appear to be here\n", argv0);
      err = NC_SHM_ERR;
      goto end;
    }
    else {
      memcpy(&curtime, &ctrl->end, sizeof(struct timeval));
      file = 1;
    }
  }

  printf("NC Information  (statistics from %s, ver %d, user %s)\nDirectory   : %s\n",
	 file ? "file" : "shared memory", ctrl->ncversion, ctrl->user, ctrl->ncdir);

  if (debugflg) {
    if (strcmp(ctrl->startdir, ctrl->ncdir))
      printf("Initial Dir : %s\n", ctrl->startdir);
    printf(".ncrc File  : %s\n", ctrl->conffile);
  }
   
  if (ctrl->start.tv_sec || ctrl->start.tv_usec) {
    /*tindex = cftime(timebuf, "%D %T", &ctrl->start.tv_sec); */
    tindex = strftime(timebuf, sizeof(timebuf), "%D %T", 
		      localtime(&ctrl->start.tv_sec));
    sprintf(timebuf + tindex, ".%04.4d", ctrl->start.tv_usec / 100);
  }

  printf("Started at  : %s, from %s", timebuf, ctrl->mastername);

  if ((ctrl->mastertype != -1) && debugflg)
    printf(" (%s)\n", typetab[ctrl->mastertype].name);
  else
    printf("\n");
      
  if (!file)
    gettimeofday(&curtime, NULL);

  cursecs = curtime.tv_sec - ctrl->start.tv_sec;
  if (ctrl->start.tv_usec > curtime.tv_usec) {
    curusecs = 1000000 - ctrl->start.tv_usec + curtime.tv_usec;
    cursecs--;
  }
  else
    curusecs = curtime.tv_usec - ctrl->start.tv_usec;
      
  sprintf(timebuf, "%02.2d:%02.2d:%02.2d.%04.4d", 
	  cursecs / (60 * 60), (cursecs / 60) % 60,
	  cursecs % 60,  curusecs / 100);
   
  printf("Elapsed time: %s\n", timebuf);

  column = printf("Environment :");
   
  for (i = 0; i < ctrl->numvars; i++) 
    if (column + strlen(vartab[i].name) + 2 >= 80)
      column = printf("\n              %s", vartab[i].name);
    else
      column += printf(" %s", vartab[i].name);
   
  printf(
	 "\n\n"
	 " Hostname     Status  F Max  Cur  Big  Bad   Tot   Load  Num  Last "
         "Start  Poll\n"
	 "                        Jobs Jobs Jobs Jobs  Jobs  Avg   CPU\n");
  curtick = ctrl->curtick;
  for(i = 0; i < ctrl->numhosts; i++) {
    if (i == ctrl->nexthost)
      ldr = "*";
    else
      ldr = " ";
    secs = hosttab[i].lastjob.tv_sec;
    usecs = hosttab[i].lastjob.tv_usec;
      
    if (secs || usecs) {
      /* tindex = cftime(timebuf, "%T", &secs); */
      tindex = strftime(timebuf, sizeof(timebuf), "%T", localtime(&secs));
      sprintf(timebuf + tindex, ".%03.3d", usecs / 1000);
    }
    else
      strcpy(timebuf, "   Never");
            
    printf("%s%-12s %-7s %c  %3d %4d %4d %4d %5d %5.2f  %2d   %-12.12s %2d", 
	   ldr, hosttab[i].name, hoststat[hosttab[i].status],
	   (hosttab[i].flags & HOST_FLAGS_FAST) ? 'F' : ' ',
	   hosttab[i].maxjobs, hosttab[i].curjobs, 
	   hosttab[i].bigjobs, hosttab[i].badjobs, hosttab[i].totjobs,
	   (double)hosttab[i].loadavg / NC_LOAD_FACTOR, hosttab[i].numcpus,
	   timebuf, 
	   (NC_POLL_INTERVAL + hosttab[i].tick - curtick) % maxticks);

    if (debugflg) {
      printf("  %-8s", ((hosttab[i].type < 0) ? 
			"Unknown" : typetab[hosttab[i].type].name));
      if (!file) 
	for (j = 0; j < hosttab[i].maxjobs; j++)
	  printf(" %5.5d", PIDGET(i, j));
    }
      
    printf("\n");
      
    cumtotjobs += hosttab[i].totjobs;
    cumbigjobs += hosttab[i].bigjobs;
    cumbadjobs += hosttab[i].badjobs;
    cumcurjobs += hosttab[i].curjobs;
    cumloadavg += hosttab[i].loadavg;

    if (secs > lastsecs || secs == lastsecs && usecs > lastusecs) {
      lastsecs = secs;
      lastusecs = usecs;
    }
      
    if (hosttab[i].status == HOST_OK)
      cumcpus += hosttab[i].numcpus;
  }

  if (lastsecs || lastusecs) {
    /* tindex = cftime(timebuf, "%T", &lastsecs); */
    tindex = strftime(timebuf, sizeof(timebuf), "%T", 
		      localtime(&lastsecs));
    sprintf(timebuf + tindex, ".%03.3d", lastusecs / 1000);
  }
  else
    strcpy(timebuf, "   Never");
   
  printf("\n  %-12s           %3d%s%4d %4d %4d %5d %5.2f  %2d   %-12.12s\n"
	 "     * - current host, F - fast host, + - jobs limited by config\n", 
	 "TOTAL", ctrl->actjobs, (ctrl->actjobs < ctrl->numjobs) ? "+" : " ",
	 cumcurjobs, cumbigjobs, cumbadjobs, cumtotjobs,
	 (double)(cumloadavg / ctrl->numhosts) / NC_LOAD_FACTOR,
	 cumcpus, timebuf);
 end:
  if (!file)
    unload_hosts(-1, 0, 0);
  else {
    free(ctrl);
    free(hosttab);
    free(vartab);
    free(typetab);
    free(transtab);
  }
   
  return err;
}

/* this function gets called to actually do a job--it runs on the same
   host as the master, but executes an rsh into a chosen remote host
   and sets up for an executes the command to 'get the job done'.
   In this case, the 'child' is the rsh command */
int nc_slave(int argc, char **argv) {
  int err = 0;

  if (load_hosts(0) == 0) {
    fprintf(stderr, "**nc** No hosts read\n");
    err = NC_SHM_ERR;
  }
  else
    err = run_job(-1, argc, argv);

  unload_hosts(child.hostnum, child.pidnum, (err != 0));
  return err;
}

/* This function gets called to 'fill in the details' about the hosts table.
   This includes getting the load averages, determining which hosts are
   actually up, and setting up the chkchildren if necessary for hosts that
   don't speak rstat (so we can poll them for their load average via the
   command line).  A host is not considered up if it doesn't have the
   appropriate directories mounted, or if it doesn't allow us to rsh in.

   Both nc_remote and nc_master use this function, in nc_master the set
   up table is provided for use by nc_slave callers via shared memory.  in
   nc_remote, the data is local to nc_remote and so is only used locally */
int complete_hosts() {
  char *nargv[4];
  int len, cmdlen;
  char *cmd, *line;
  int notdone = 1;
  int ready_fds, num_fds = 0, max_fd = 0;
  int i, j;
  struct timeval timeout;
  fd_set rfds;
   
  FD_ZERO(&rfds);

  if (!(chkchildren = (struct child *)malloc(sizeof(struct child) * 
					     (ctrl->numhosts)))) {
    fprintf(stderr, "No memory for check children\n");
    return NC_RES_ERR;
  }
   
  cmdlen = sizeof(NC_CD) + strlen(pwd) + sizeof(NC_ERR_REQ) + 3;
  cmd = alloca(cmdlen);
  sprintf(cmd, NC_CD " %s; " NC_ERR_REQ "\n", pwd);
   
  nargv[0] = NC_RSH;
  nargv[2] = NC_SH;
  nargv[3] = NULL;
   
   
   
  /* initialize all of the chkchildren structures.  
     children are used to check for the existence of
     directories on the remote */
  for (i = 0; i < ctrl->numhosts; i++) {
    chkchildren[i].out.eof = 0;
    chkchildren[i].out.ptr = chkchildren[i].out.buf;
    chkchildren[i].out.bytes = 0;
    chkchildren[i].err.eof = 0;
    chkchildren[i].err.ptr = chkchildren[i].err.buf;
    chkchildren[i].err.bytes = 0;
    chkchildren[i].errcode = 0;
    chkchildren[i].env = NULL;
    chkchildren[i].argv = NULL;
    chkchildren[i].argc = 0;
      
    chkchildren[i].flags = STDOUT | STDERR | STDIN;
    chkchildren[i].hostnum = i;
    chkchildren[i].state = NC_STATE_CD_WAIT;
      
    nargv[1] = hosttab[i].name;
      
    if (j = spawn_child(nargv, &(chkchildren[i]))) {
      fprintf(stderr, "**nc** Error spawning on %s %d\n",
	      hosttab[i].name, j);
      return NC_GEN_ERR;
    }
    debug(2, "master sending %s command: %s", 
	  hosttab[i].name, cmd);
    write(chkchildren[i].sin, cmd, strlen(cmd));
    FD_SET(chkchildren[i].sout, &rfds);
    FD_SET(chkchildren[i].serr, &rfds);
    num_fds += 2;
    if (max_fd < chkchildren[i].sout || max_fd < chkchildren[i].serr)
      max_fd = max(chkchildren[i].sout, chkchildren[i].serr);
  }
   
  /* notdone tracks if there are still any poll children that
     haven't responded yet.  No since in waiting for an extra
     10 seconds if everyone is up */
  while (notdone) {
    /* wait for the poll interval on all of the polling children.
       This loop actually waits for 5 seconds after the last
       response, unless everyone has already responded.  It probably
       should just wait for a cumulative 5 seconds instead */
    timeout.tv_sec = NC_POLL_WAIT;
    timeout.tv_usec = 0;

    ready_fds = select(max_fd + 1, &rfds, NULL, NULL, &timeout);
    if (!ready_fds)
      break;
    else if (ready_fds < 0) {
      printf("ready_fds %d--errno %d\n", ready_fds, errno);
      break;
    }
	
    notdone = 0;

    for(i = 0; i < ctrl->numhosts; i++) {
      if (chkchildren[i].pid) {
	if (FD_ISSET(chkchildren[i].sout, &rfds)) {
          int more;
          do if (line = get_line(&(chkchildren[i]), STDOUT, &len, &more))
            process_master(&(chkchildren[i]), line, len, 
                           MASTER_NOT_QUIET, 0);
          while (more);
	  ready_fds--;
	}
	if (FD_ISSET(chkchildren[i].serr, &rfds)) {
          int more;
          do 
            if (line = get_line(&(chkchildren[i]), STDERR, &len, &more))
              fprintf(stderr, "%s:%s\n", hosttab[i].name, line);
          while (more);
	  ready_fds--;
	}
	if (chkchildren[i].err.eof && chkchildren[i].out.eof) {
	  hosttab[i].status = HOST_DEAD;
	  kill_child(&(chkchildren[i]));
	}
      }
      FD_CLR(chkchildren[i].sout, &rfds);
      FD_CLR(chkchildren[i].serr, &rfds);
         
      if (hosttab[i].status == HOST_UNKNOWN) {
	/* if it is unknown, then we haven't heard yet,
	   keep listening */
	FD_SET(chkchildren[i].sout, &rfds);
	FD_SET(chkchildren[i].serr, &rfds);
	notdone = 1;
      }
    }
    if (ready_fds)
      fprintf(stderr, "**nc** missed %d file descriptors\n", ready_fds);
  }
   
  /* now, for any hosts that we haven't heard from, we kill them off,
     Any still living children will be used for load average computation,
     as they don't appear to support rstat */
  for (i = 0; i < ctrl->numhosts; i++)
    if (hosttab[i].status == HOST_UNKNOWN) {
      hosttab[i].status = HOST_DEAD;
      fprintf(stderr, "couldn't contact %s\n", hosttab[i].name);
      kill_child(&(chkchildren[i]));
    }
  
  if (!marg && ctrl->numtrans) {
    short host;
    struct utsname systeminfo;

    if (uname(&systeminfo) < 0) {
      fprintf(stderr, "Call to uname() failed\n");
      return NC_RES_ERR;
    }

    if (((host = find_host(ctrl->numhosts, systeminfo.nodename)) >= 0) &&
	(hosttab[host].type != -1)) {
      debug(1, "getting host type via nodename\n");
      ctrl->mastertype = hosttab[host].type;
    }
    else {
      debug(1, "getting host type via sysname\n");
      if ((ctrl->mastertype = find_type(ctrl->numtypes, 
					systeminfo.sysname)) < 0) {
	fprintf(stderr, "Host type %s not found\n", systeminfo.sysname);
	return NC_NCRC_ERR;
      }
    }
  }
  return 0;
}

int fix_remote_cmdline(int argc, char **argv) {
  int i;
  int j = remoteflg;

  if (!remoteflg)
    return 1;

  /* set up command line arguments for remote nc */
  argv[0] = progname;
  /* filter out any -r arguments */
  for (i = 1; i <= argc; i++) {
    char *pos;
    if (argv[i][0] != '-')
      continue;
    pos = strchr(argv[i], 'r');
    if (pos) {
      *pos = 'x';
      if (!--j)
        break;
    }
  } 
  return j;
}

/* This function actually runs a job specified by the provided argc,argv on a
   remote machine, which will be picked by pick_host */
int run_job(int hostnum, int argc, char **argv) {
  fd_set rfds;
  struct timeval timeout;
  int ready_fds, num_fds, max_fd;
  char *nargv[4];
  char **args;
  char *line;
  int len;
  int cmdlen;
  char *cmd;

  FD_ZERO(&rfds);

  child.out.eof = 0;
  child.out.ptr = child.out.buf;
  child.out.bytes = 0;
  child.argv = argv;
  child.argc = argc;
  child.err.eof = 0;
  child.err.ptr = child.err.buf;
  child.err.bytes = 0;
   
  child.flags = STDOUT | STDERR | STDIN;
  child.errcode = 0;
  child.env = NULL;
  child.pidnum = -1;

  if (hostnum != -1)
    child.hostnum = hostnum;
  else if ((child.hostnum = pick_host(&child.pidnum, bigflg)) == -1) {
    fprintf(stderr, "**nc** Failed to pick a host\n");
    return NC_GEN_ERR;
  }
   
  nargv[0] = NC_RSH;
  nargv[1] = hosttab[child.hostnum].name;
  nargv[2] = NC_SH;
  nargv[3] = NULL;
   
  gettimeofday(&hosttab[child.hostnum].lastjob, NULL);
   
  if (spawn_child(nargv, &child)) {
    fprintf(stderr, "**nc** Error spawning child %s\n", nargv[1]);
    child.hostnum = -1;
    return NC_GEN_ERR;
  }
  if (child.pidnum != -1)
    PIDSET(child.hostnum, child.pidnum, child.pid);
  cmdlen = sizeof(NC_CD) + strlen(pwd) + sizeof(NC_STAT_REQ) + 3;
  cmd = alloca(cmdlen);
  sprintf(cmd, NC_CD " %s; " NC_STAT_REQ "\n", pwd);
  debug(2, "sending %s command: %s", hosttab[child.hostnum].name, cmd);
  write(child.sin, cmd, strlen(cmd));
   
  child.state = NC_STATE_CD_WAIT;
  max_fd = max(child.sout, child.serr);
   
  FD_SET(child.sout, &rfds);
  FD_SET(child.serr, &rfds);

  while (!child.out.eof || !child.err.eof) {
    timeout.tv_sec = NC_POLL_WAIT * 2;
    timeout.tv_usec = 0;

    ready_fds = select(max_fd + 1, &rfds, NULL, NULL, &timeout);
    if (ready_fds == 0) {
      /* timeout */
      if (hosttab[child.hostnum].status == HOST_DEAD) {
	fprintf(stderr, "**nc** Host %s died\n",
		hosttab[child.hostnum].name);
	return NC_TIMEOUT;
      }
    }
    else if (ready_fds < 0) {
      fprintf(stderr, "**nc**: %s Call to select() failed", strerror(errno));
      return NC_GEN_ERR;
    }

    if (FD_ISSET(child.sout, &rfds)) {
      int more;
      do 
        if (line = get_line(&child, STDOUT, &len, &more))
          process_slave(&child, line, len);
      while (more);
      ready_fds--;
    }
    else if (!child.out.eof)
      FD_SET(child.sout, &rfds);
      
    if (FD_ISSET(child.serr, &rfds)) {
      int more;
      do 
        if (line = get_line(&child, STDERR, &len, &more))
          if (child.state == NC_STATE_CMD_WAIT)
            fprintf(stderr, "%s:%s\n", hosttab[child.hostnum].name, line);
          else {
            fprintf(stderr, "**nc** Got error in state %d from %s: %s\n",
                    child.state, hosttab[child.hostnum].name, line);
            /* start avoiding the host immediately ***TODO*** */
            hosttab[child.hostnum].status = HOST_DEAD;
            return NC_TIMEOUT;
          }
      while (more);
      ready_fds--;
    }
    else if (!child.err.eof)
      FD_SET(child.serr, &rfds);

    if (ready_fds) {
      fprintf(stderr, "**nc** Select problem\n");
      return NC_GEN_ERR;
    }
  }
  return child.errcode;
}

/* run nc on a remote machine, picking the machine just as we would to
   run a slave--nc is considered a 'big job' for this purpose.

   or run 'nc -c' or 'nc -i' on a remote machine specified by a .ncmaster
   file */
int nc_remote(int argc, char **argv) {
  int err = 0;
  int hostnum = -1;
  int i;

  if (!create_hosts()) {
    fprintf(stderr, "No hosts read\n");
    err = NC_NCRC_ERR;
    goto end;
  }

  if (fix_remote_cmdline(argc, argv)) {
    err = NC_GEN_ERR;
    goto end;
  }
  if (infoflg || cleanflg) {
    hostnum = read_master();
    if (hostnum == -1) {
      fprintf(stderr, "failed to read master\n");
      err = NC_GEN_ERR;
      goto end;
    }
    /* fake out this one host as dead in order to
       allow the timeout to make us exit */
    hosttab[hostnum].status = HOST_DEAD;
  }
  else {
    FILE *masterfile;
    int pidnum;
    if ((err = complete_hosts()))
      goto end;
    hostnum = pick_host(&pidnum, 1);
    if (!(masterfile = fopen(ncmasterfile, "w")))
	fprintf(stderr, "Couldn't save master\n");
      else {
        fwrite(hosttab[hostnum].name, strlen(hosttab[hostnum].name), 1,
               masterfile);
	fclose(masterfile);
      }
  }
  err = run_job(hostnum, argc, argv);
  /* don't need to unload_hosts,
     since we don't have any shared memory */
 end:
  return err;
}

int nc_master(int argc, char **argv) {
  int err = 0;
  fd_set rfds;
  int ready_fds, num_fds = 0, max_fd = 0;
  char **args;
  int cmdlen, len;
  char *cmd, *line;
  long basetime;            /* the tv_sec field that is the base for
			       the current tick count */
  int newtick = 0;          /* new tick counter */
  struct timeval timeout;
  char jobstr[NC_MAX_DIGITS + 1];
  char *env;
  int possible_jobs = 0;
  int i, j;
  int delta = maxticks - 1;
   
  FD_ZERO(&rfds);

  if (!create_hosts()) {
    fprintf(stderr, "No hosts read\n");
    err = NC_NCRC_ERR;
    goto real_end;
  }

  if ((err = complete_hosts()))
    goto end;
  
  /* now, for any hosts that are OK, we add their jobs in, 
     and initialize their ticker to a random value.  Those
     children should already be in the sleep state, as desired */
  for (i = 0; i < ctrl->numhosts; i++) {
    if (hosttab[i].status == HOST_OK) {
      possible_jobs += hosttab[i].maxjobs;
      hosttab[i].tick = rand() % (maxticks - 1);
      if (delta > hosttab[i].tick)
        delta = hosttab[i].tick;

      if (chkchildren[i].pid) {
	FD_SET(chkchildren[i].sout, &rfds);
	FD_SET(chkchildren[i].serr, &rfds);
      }
    }
  }
  if (ctrl->actjobs + margin > possible_jobs)
    if (possible_jobs > margin)
      ctrl->actjobs = possible_jobs - margin;
    else if (possible_jobs > NC_DEF_MARGIN)
      ctrl->actjobs = possible_jobs - NC_DEF_MARGIN;
    else
      ctrl->actjobs = possible_jobs;

  if (ctrl->actjobs > 0) { 
    /* need to set the environment variables that
       will make 'make' work properly */
       
    putenv(NC_BIG_ENV "=" NC_BIG_VAL);
    putenv(NC_QUIET_ENV "=" NC_QUIET_VAL);
       
    env = malloc(strlen(progname) + sizeof(NC_PROG_ENV "=%s -s"));
    sprintf(env, NC_PROG_ENV "=%s -s", progname);
    cmdstr = strchr(env, '=') + 1;
    cmdstrlen = strlen(cmdstr);
    putenv(env);

    env = malloc(strlen(progname) + sizeof(NC_MASTER_ENV "=%s"));
    sprintf(env, NC_MASTER_ENV "=%s", progname);
    putenv(env);

    sprintf(jobstr, "%d", ctrl->actjobs);

    env = malloc(sizeof(NC_JOBS_ENV) + NC_MAX_DIGITS + 2);
    sprintf(env, NC_JOBS_ENV "=%d", ctrl->actjobs);
    putenv(env);

    /* we need argc arguments that we passed on the
       command line, plus 2 for the -j and the number of jobs, plus
       one for the trailing NULL */
    args = malloc(sizeof(char *) * (argc + 3));
    args[0] = argv[0];
    /* rather bogus, we assume that whatever we're invoking uses a 
       "-j N" parameter, to say how many jobs to run */
    args[1] = NC_JOB_VAL;
    args[2] = jobstr;
   
    /* copy the command line arguments */
    for (i = 1; i < argc; i++)
      args[i + 2] = argv[i];
       
    args[i + 2] = NULL;
  } else {
    /* Strange, but not fatal--we haven't anyplace to run jobs
       Just issue warning and run make */
    fprintf(stderr, "%s : disabling distributed mode\n", argv[0]);
    args = malloc(sizeof(char *) * (argc+1));
    for (i = 0; i < argc; i++)
      args[i] = argv[i];
    args[i] = NULL;
    cmdstr = NULL;
    cmdstrlen = 0;
  }
   
  /* display what command we're invoking */
  printf("Invoking:");
  for (i = 0; args[i]; i++)
    printf(" %s", args[i]);
  printf("\n");
   
  child.out.eof = 0;
  child.out.ptr = child.out.buf;
  child.out.bytes = 0;
  child.err.eof = 0;
  child.err.ptr = child.err.buf;
  child.err.bytes = 0;
  child.errcode = 0;
  child.env = NULL;
  child.argv = NULL;
  child.argc = 0;
  child.flags = STDOUT | STDERR;
   
  /* now we can finally spawn the 'make' child */
  if (spawn_child(args, &child)) {
    fprintf(stderr, "**nc** Error spawning program %s\n", args[0]);
    err = NC_GEN_ERR;
    goto end;
  }
   
  FD_SET(child.sout, &rfds);
  FD_SET(child.serr, &rfds);
  num_fds += 2;
  if (max_fd < child.sout || max_fd < child.serr)
    max_fd = max(child.sout, child.serr);
   
  /* initialize our ticker */
  gettimeofday(&curtime, NULL);
  basetime = curtime.tv_sec;

  /* watch all children fds, as well as timers. keep doing this until
     the build is done */   
  do {
    /* set the timeout to be the current delta until the
       next child times out */
    timeout.tv_sec = delta;
    timeout.tv_usec = 0;
    
    ready_fds = select(max_fd + 1, &rfds, NULL, NULL, &timeout);
     
    /* determine how many seconds have passed since select was
       called, and set newtick appropriately */
    gettimeofday(&curtime, NULL);
    newtick = (curtime.tv_sec - basetime + ctrl->curtick) % maxticks;
      
    /* start resetting the delta, based on the 'poll' children */
    delta = maxticks - 1;
      
    for (i = 0; i < ctrl->numhosts; i++) {
      /* determine if any timers have expired, 
	 handle the timer, 
	 then check for stream activity */
      if (hosttab[i].status == HOST_DEAD ||
	  hosttab[i].status == HOST_AVOID)
	continue;
         
      if ((newtick < ctrl->curtick && 
	   (hosttab[i].tick > ctrl->curtick || 
	    hosttab[i].tick <= newtick)) ||
	  (newtick > ctrl->curtick && 
	   (hosttab[i].tick > ctrl->curtick && 
	    hosttab[i].tick <= newtick))) {
	/* timer did expire */
#ifdef RSTAT
	/* chkchildren[i] will have a pid only if the host doesn't
	   support the rstat call--in which case we use an interactive
	   call to determine the current load average on the */
	if (!chkchildren[i].pid) {
	  struct statstime stat;
	  if (rstat(hosttab[i].name, &stat) == RPC_SUCCESS) {
	    debug(2, "got rstat from %s\n", hosttab[i].name);
	    hosttab[i].status = HOST_OK;
	    hosttab[i].loadavg = stat.avenrun[0];
	    hosttab[i].tick = (hosttab[i].tick + NC_POLL_INTERVAL) % 
	      maxticks;
	  }
	  else {
	    /* we got no response.  At first, move him to failing, then
	       wait again.  If still no response, then kill him */
	    if (hosttab[i].status == HOST_FAILING) {
	      hosttab[i].status = HOST_DEAD;
	      kill_slaves(ALLPIDS(i), hosttab[i].maxjobs);
	      fprintf(stderr, "%s didn't respond to rstat\n",
		      hosttab[i].name);
	    }
	    else {
	      hosttab[i].status = HOST_FAILING;
	      hosttab[i].tick = (hosttab[i].tick + NC_POLL_WAIT) % 
		maxticks;
	    }
	  }
	}
	else
#endif
	  if (chkchildren[i].state == NC_STATE_SLEEP) {
	    /* if he's been sleeping, time to send the
	       uptime command again. */
	    cmdlen = sizeof(NC_UPTIME) + 1;
	    cmd = alloca(cmdlen);
	    strcpy(cmd, NC_UPTIME "\n");
	    debug(2, "master sending %s command: %s", 
		  hosttab[i].name, cmd);
	    write(chkchildren[i].sin, cmd, strlen(cmd));
                  
	    /* reset his ticker */
	    chkchildren[i].state = NC_STATE_LOAD_WAIT;
	    hosttab[i].tick = (hosttab[i].tick + NC_POLL_WAIT) % 
	      maxticks;
	  }
	  else {
	    /* he wasn't sleeping, but we got no response.
	       At first, move him to failing, then
	       wait again.  If still no response, then kill him */
	    if (hosttab[i].status == HOST_FAILING) {
	      hosttab[i].status = HOST_DEAD;
	      fprintf(stderr, "%s didn't respond\n",
		      hosttab[i].name);
	      kill_slaves(ALLPIDS(i), hosttab[i].maxjobs);
	      kill_child(&(chkchildren[i]));
	    }
	    else {
	      hosttab[i].status = HOST_FAILING;
	      hosttab[i].tick = (hosttab[i].tick + NC_POLL_WAIT) % 
		maxticks;
	    }
	  }
      }

      /* see if chkchildren[i] has any I/O to process */
      if (chkchildren[i].pid) {
	if (FD_ISSET(chkchildren[i].sout, &rfds)) {
          int more;
          do 
            if (line = get_line(&(chkchildren[i]), STDOUT, &len, &more))
              process_master(&(chkchildren[i]), line, len, MASTER_QUIET,
                             newtick);
          while (more);
	  ready_fds--;
	}
	else
	  /* must reset fd bit for next pass */
	  FD_SET(chkchildren[i].sout, &rfds);
            
	if (FD_ISSET(chkchildren[i].serr, &rfds)) {
          int more;
	  do 
            if (line = get_line(&(chkchildren[i]), STDERR, &len, &more))
              fprintf(stderr, "%s\n", line);
          while (more);
	  ready_fds--;
	}
	else 
	  /* must reset fd bit for next pass */
	  FD_SET(chkchildren[i].serr, &rfds);

	/* eof conditions mean that we've had a network interruption,
	   we can no longer count on using this host */
	if (chkchildren[i].err.eof && chkchildren[i].out.eof) {
          /* should mark host as dead if _either_ are eof??? ***TODO*** */
	  hosttab[chkchildren[i].hostnum].status = HOST_DEAD;
	  kill_slaves(ALLPIDS(i), hosttab[i].maxjobs);
	  kill_child(&(chkchildren[i]));
	}
      }
         
      /* find out what the new delta should be */
      if (hosttab[i].status != HOST_DEAD &&
	  hosttab[i].status != HOST_AVOID)
	if (hosttab[i].tick > newtick &&
	    delta > hosttab[i].tick - newtick)
	  delta = hosttab[i].tick - newtick;
	else if (delta > maxticks - (newtick - hosttab[i].tick))
	  delta = maxticks - (newtick - hosttab[i].tick);
         
    }

    /* read the output from the 'make' child, if any */
      
    if (FD_ISSET(child.sout, &rfds)) {
      int more;
      do 
        if (line = get_line(&(child), STDOUT, &len, &more))
          process_cmd(&(child), STDOUT, line, len);
      while (more);
      ready_fds--;
    }
    else
      FD_SET(child.sout, &rfds);
      
    if (FD_ISSET(child.serr, &rfds)) {
      int more;
      do 
        if (line = get_line(&(child), STDERR, &len, &more)) {
          debug(3, "err line was: '%s'\n", line);
          fprintf(stderr, "%s\n", line);
        }
      while (more);
      ready_fds--;
    }
    else
      FD_SET(child.serr, &rfds);
      
    if (ready_fds)
      fprintf(stderr, "**nc** missed %d file descriptors\n", ready_fds);
      
    ctrl->curtick = newtick;
    basetime = curtime.tv_sec;
  } while(!child.out.eof || !child.err.eof);

  /* Have to check child return code now */
  if (waitpid(child.pid, &child.exitcode, 0) == -1){
    err = 1;
    fprintf(stderr, "**nc** %s Cannot get spawned child return code", 
	    strerror(errno));
  } else {
    if (WIFEXITED(child.exitcode)){
      err = WEXITSTATUS(child.exitcode);
      /* 	   fprintf(stderr, "**nc** Child exited with code %d\n", err); */
    } else if (WIFSIGNALED(child.exitcode)){
      fprintf(stderr, "**nc** Child killed with signal %d\n", 
	      WTERMSIG(child.exitcode));
      if (WCOREDUMP(child.exitcode))
	fprintf(stderr, "**nc** Child dumped core\n");
      err = 1;
    } else {
      fprintf(stderr, "**nc** Child exited in a strange way %08X\n", 
	      child.exitcode);
      err = 1;
    }
  }
   
 end:
  unload_hosts(-1, 0, 0);
 real_end:
  return err;
}
   
int main(int argc, char **argv, char **env) {
  char opt;
  struct sigaction action;
  
  argv0 = argv[0];

  /* line buffer both stdout and stderr */
  setvbuf(stdout, NULL, _IOLBF, 0);
  setvbuf(stderr, NULL, _IOLBF, 0);

  child.hostnum = -1;
   
  while ((opt = getopt(argc, argv, "bcdfij:k:mp:qrst:vx")) != EOF) {
    switch(opt) {
    case 'b':
      bigflg++;
      break;
    case 'c':
      cleanflg++;
      break;
    case 'd':
      debugflg++;
      break;
    case 'f':
      fileflg++;
      break;
    case 'i':
      infoflg++;
      break;
    case 'j':
      jobflg++;
      if (sscanf(optarg, "%d", &jobarg) != 1 ||
	  jobarg < 1 || jobarg > NC_MAX_JOBS)
	errflg++;
      break;
    case 'k':         
      if (sscanf(optarg, "%x,%x", &ctrlkey, &tabkey) != 2)
	errflg++;
      break;
    case 'm':
      memflg++;
      break;
    case 'p':
      pathflg++;
      parg = optarg;
      break;
    case 'q':
      quietflg++;
      break;
    case 'r':
      remoteflg++;
      break;
    case 's':
      slaveflg++;
      break;
    case 't':
      marg = optarg;
      break;
    case 'v':
      printf("%s version: %d. %s\n", argv[0], NC_VERSION, nc_cvsid);
      break;
    case 'x':
      xflg++;
      break;
    default:
      errflg++;
    }
      
    if (errflg)
      goto usage;
  }
   
  if ((optind == argc && (slaveflg || !(cleanflg || infoflg))) ||
      (optind < argc && (cleanflg || infoflg)) ||
      ((fileflg || memflg) && !(cleanflg || infoflg)) ||
      (fileflg && memflg) ||
      ((bigflg  || quietflg) && !slaveflg) ||
      ((pathflg || marg || remoteflg) && slaveflg)) {
  usage:
    fprintf(stderr, 
	    "usage: %1$s [-r][-p <ncdir>][-j <maxjobs>][-t <type>] <make-cmd>\n"
	    "       %1$s [-s [-b][-q]] <compile-cmd>\n"
	    "       %1$s [-i|-c][-f|-m][-r][-p <ncdir>]\n\n"
	    "<make-cmd> is normally \"make\" plus other args\n"
	    "<compile-cmd> is any command that can be invoked via \"make\"\n"
            "r - remote nc, p - use <ncdir> as path for finding nc data\n"
	    "j - no more than <maxjobs> jobs, t - master is a <type> host\n"
	    "s - nc slave, b - big job, q - quiet\n"
	    "i - nc status information, c - cleanup shared memory\n"
	    "f - nc data from file only, m - nc data from shared memory only\n",
	    argv[0]);
    exit(NC_USAGE_ERR);
  }
   
  if (exec_path(argv0, progname))
    exit(NC_PATH_ERR);

  getcwd(pwd, FILENAME_MAX);

  if (slaveflg)
    if (!(ncdir = getenv(NC_DIR_ENV))) {
      fprintf(stderr, "**nc** " NC_DIR_ENV " not defined\n");
      exit(NC_PATH_ERR);
    }
    else
      debug(1, "Got NC path from environment: %s\n", ncdir);
  else {
    /* must set NC_DIR environment variable if not already set
       if set, must make it absolute */
    char tmpdir[FILENAME_MAX + 1];
    char *pdir = tmpdir;
    char *var;

    if (parg) {
      if (absolute_path(parg, pwd, tmpdir))
	exit(NC_PATH_ERR);
      debug(1, "absolute path: %s\n", tmpdir);
    }
    else if (ncdir) {
      if (absolute_path(ncdir, pwd, tmpdir))
	exit(NC_PATH_ERR);
      debug(1, "absolute " NC_DIR_ENV ": %s\n", tmpdir);
    }
    else
      pdir = pwd;
     
    var = malloc(sizeof(NC_DIR_ENV "=%s") + strlen(pdir));
    sprintf(var, NC_DIR_ENV "=%s", pdir);
    putenv(var);
    debug(1, "putenv: %s\n", var);

    ncdir = getenv(NC_DIR_ENV);
  }
   

  sprintf(ncstatfile, "%s/" NC_STATFILE, ncdir);
  sprintf(ncmasterfile, "%s/" NC_MASTERFILE, ncdir);
#if 0   
  sigset(SIGHUP, handle_sig);
  sigset(SIGINT, handle_sig);
  sigset(SIGQUIT, handle_sig);
  sigset(SIGTERM, handle_sig);
  sigset(SIGXCPU, handle_sig);
#if 0
  sigset(SIGILL, handle_sig);
  sigset(SIGBUS, handle_sig);
  sigset(SIGSEGV, handle_sig);
  sigset(SIGSYS, handle_sig);
#endif
#else
  action.sa_handler = handle_sig;
  sigemptyset(&action.sa_mask);
  action.sa_flags = SA_NOCLDWAIT;

  sigaction(SIGHUP, &action, NULL);
  sigaction(SIGINT, &action, NULL);
  sigaction(SIGQUIT, &action, NULL);
  sigaction(SIGTERM, &action, NULL);
  sigaction(SIGXCPU, &action, NULL);
#if 0
  sigaction(SIGILL, &action, NULL);
  sigaction(SIGBUS, &action, NULL);
  sigaction(SIGSEGV, &action, NULL);
  sigaction(SIGSYS, &action, NULL);
#endif
#endif
   
  /* need to initialize the random number generator */
  gettimeofday(&curtime, NULL);
  srand((unsigned int)(curtime.tv_sec ^ curtime.tv_sec ^ (int)getpid()));

  if (slaveflg)
    return nc_slave(argc - optind, argv + optind);
  else if (remoteflg)
    return nc_remote(argc, argv);
  else if (cleanflg) 
    return nc_clean();
  else if (infoflg)
    return nc_info();
  else
    return nc_master(argc - optind, argv + optind);
}
